import numpy as np
import pandas as pd
import pytest

from cuanalytics import fit_logit
from cuanalytics.regression.logistic import LogisticRegressionModel


@pytest.fixture
def binary_data():
    np.random.seed(42)
    x1 = np.random.normal(0, 1, 100)
    x2 = np.random.normal(0, 1, 100)
    logits = 0.5 * x1 - 0.3 * x2
    probs = 1 / (1 + np.exp(-logits))
    y = np.where(probs > 0.5, 'A', 'B')
    return pd.DataFrame({'x1': x1, 'x2': x2, 'class': y})


def test_fit_logit_returns_model(binary_data):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    assert isinstance(model, LogisticRegressionModel)

def test_missing_formula_raises(binary_data):
    with pytest.raises(ValueError, match="Must provide 'formula'"):
        fit_logit(binary_data, formula=None)


def test_model_missing_formula_raises(binary_data):
    with pytest.raises(ValueError, match="Must provide 'formula'"):
        LogisticRegressionModel(binary_data, formula=None)


def test_invalid_formula_raises(binary_data):
    with pytest.raises(ValueError, match="Formula must include a target"):
        fit_logit(binary_data, formula='class x1 + x2')


def test_missing_target_column_raises(binary_data):
    with pytest.raises(ValueError, match="Target 'missing' not found"):
        fit_logit(binary_data, formula='missing ~ x1 + x2')


def test_formulaic_missing_raises(binary_data, monkeypatch):
    import builtins
    original_import = builtins.__import__

    def mock_import(name, *args, **kwargs):
        if name == 'formulaic':
            raise ImportError("No module named 'formulaic'")
        return original_import(name, *args, **kwargs)

    monkeypatch.setattr(builtins, '__import__', mock_import)
    with pytest.raises(ImportError, match="requires the 'formulaic' library"):
        fit_logit(binary_data, formula='class ~ x1 + x2')


def test_non_numeric_feature_raises(binary_data):
    df = binary_data.copy()
    df['cat'] = ['a', 'b'] * 50
    with pytest.raises(ValueError, match="All features must be numeric"):
        fit_logit(df, formula='class ~ x1 + cat')


def test_categorical_feature_with_C_allows_fit(binary_data):
    df = binary_data.copy()
    df['cat'] = ['a', 'b'] * 50
    model = fit_logit(df, formula='class ~ x1 + C(cat)')
    assert isinstance(model, LogisticRegressionModel)


def test_unfitted_predict_raises(binary_data):
    model = LogisticRegressionModel.__new__(LogisticRegressionModel)
    with pytest.raises(RuntimeError, match="not been fitted"):
        model.predict(binary_data)


def test_predict_returns_series(binary_data):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    preds = model.predict(binary_data)
    assert isinstance(preds, pd.Series)
    assert len(preds) == len(binary_data)


def test_score_returns_report(binary_data, monkeypatch):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
    report = model.score(binary_data)
    assert isinstance(report, dict)
    assert 'accuracy' in report
    assert 'confusion_matrix' in report


def test_get_metrics(binary_data):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    metrics = model.get_metrics()
    assert metrics['model_type'] == 'logistic_regression'
    assert 'coefficients' in metrics


def test_summary_returns_none(binary_data, monkeypatch):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
    result = model.summary()
    assert result is None


def test_predict_proba(binary_data):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    probs = model.predict_proba(binary_data)
    assert probs.shape[0] == len(binary_data)
    assert probs.shape[1] == len(model.classes)


def test_random_state_passthrough(binary_data):
    model = fit_logit(
        binary_data,
        formula='class ~ x1 + x2',
        solver='liblinear',
        random_state=123,
    )
    assert model.random_state == 123
    assert model.model.random_state == 123


def test_transform_missing_model_spec_raises(binary_data):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    model.model_spec = None
    with pytest.raises(RuntimeError, match="Formula metadata missing"):
        model._transform_data_with_formula(binary_data)


def test_transform_rhs_attribute_branch(binary_data, monkeypatch):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')

    class DummyMatrices:
        def __init__(self, rhs):
            self.rhs = rhs

    class DummySpec:
        def get_model_matrix(self, data, output='pandas'):
            rhs = pd.DataFrame(
                {
                    'Intercept': [1] * len(data),
                    'x1': data['x1'].values,
                    'x2': data['x2'].values
                },
                index=data.index
            )
            return DummyMatrices(rhs)

    model.model_spec = DummySpec()
    X_new = model._transform_data_with_formula(binary_data)
    assert list(X_new.columns) == ['x1', 'x2']


def test_init_rhs_attribute_branch(binary_data, monkeypatch):
    class DummySpec:
        def __init__(self):
            self.variables = set()

    class DummyMatrices:
        def __init__(self, rhs):
            self.rhs = rhs
            self.model_spec = DummySpec()

    def mock_model_matrix(formula, data, output='pandas'):
        rhs = pd.DataFrame(
            {
                'Intercept': [1] * len(data),
                'x1': data['x1'].values,
                'x2': data['x2'].values
            },
            index=data.index
        )
        return DummyMatrices(rhs)

    import formulaic
    monkeypatch.setattr(formulaic, 'model_matrix', mock_model_matrix)

    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    assert model.feature_names == ['x1', 'x2']


def test_visualize_runs(binary_data, monkeypatch):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
    model.visualize()


def test_visualize_multiclass_coefficients(monkeypatch):
    np.random.seed(7)
    x1 = np.random.normal(0, 1, 90)
    x2 = np.random.normal(0, 1, 90)
    y = np.array(['A'] * 30 + ['B'] * 30 + ['C'] * 30)
    df = pd.DataFrame({'x1': x1, 'x2': x2, 'class': y})
    model = fit_logit(df, formula='class ~ x1 + x2')
    monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
    model.visualize()


def test_visualize_features_runs(binary_data, monkeypatch):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
    model.visualize_features('x1', 'x2')


def test_visualize_features_invalid_feature(binary_data):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    with pytest.raises(ValueError, match="Features must be from"):
        model.visualize_features('x1', 'missing')


@pytest.mark.filterwarnings("ignore:.*penalty.*:FutureWarning")
@pytest.mark.filterwarnings("ignore:.*penalty.*:UserWarning")
def test_visualize_features_penalty_non_default(binary_data, monkeypatch):
    model = fit_logit(binary_data, formula='class ~ x1 + x2', penalty='l1', solver='liblinear')
    monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
    model.visualize_features('x1', 'x2')


def test_visualize_all_features_runs(binary_data, monkeypatch):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
    model.visualize_all_features(cols=1)


def test_visualize_all_features_single_axis(monkeypatch):
    np.random.seed(11)
    x1 = np.random.normal(0, 1, 50)
    logits = 0.8 * x1
    probs = 1 / (1 + np.exp(-logits))
    y = np.where(probs > 0.5, 'A', 'B')
    df = pd.DataFrame({'x1': x1, 'class': y})
    model = fit_logit(df, formula='class ~ x1')
    monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
    model.visualize_all_features(cols=1)


def test_visualize_all_features_hides_unused_axes(binary_data, monkeypatch):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
    model.visualize_all_features(cols=3)


def test_visualize_all_features_multiclass_error(monkeypatch):
    np.random.seed(8)
    x1 = np.random.normal(0, 1, 90)
    x2 = np.random.normal(0, 1, 90)
    y = np.array(['A'] * 30 + ['B'] * 30 + ['C'] * 30)
    df = pd.DataFrame({'x1': x1, 'x2': x2, 'class': y})
    model = fit_logit(df, formula='class ~ x1 + x2')
    with pytest.raises(ValueError, match="only supported for binary"):
        model.visualize_all_features()


def test_coeff_table_handles_singular_matrix(binary_data, monkeypatch):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')

    def mock_inv(*args, **kwargs):
        raise np.linalg.LinAlgError("singular")

    monkeypatch.setattr(np.linalg, 'inv', mock_inv)
    table = model.get_coefficients()
    assert table['z_score'].isna().any()
    assert table['significance'].isin(['', '***', '**', '*', '.']).all()


def test_coeff_table_multiclass_branch(monkeypatch):
    np.random.seed(9)
    x1 = np.random.normal(0, 1, 90)
    x2 = np.random.normal(0, 1, 90)
    y = np.array(['A'] * 30 + ['B'] * 30 + ['C'] * 30)
    df = pd.DataFrame({'x1': x1, 'x2': x2, 'class': y})
    model = fit_logit(df, formula='class ~ x1 + x2')
    table = model.get_coefficients()
    assert set(table['class']) == set(model.classes)


def test_coeff_table_multiclass_singular(monkeypatch):
    np.random.seed(10)
    x1 = np.random.normal(0, 1, 90)
    x2 = np.random.normal(0, 1, 90)
    y = np.array(['A'] * 30 + ['B'] * 30 + ['C'] * 30)
    df = pd.DataFrame({'x1': x1, 'x2': x2, 'class': y})
    model = fit_logit(df, formula='class ~ x1 + x2')

    def mock_inv(*args, **kwargs):
        raise np.linalg.LinAlgError("singular")

    monkeypatch.setattr(np.linalg, 'inv', mock_inv)
    table = model.get_coefficients()
    assert table['z_score'].isna().any()


def test_significance_stars(binary_data):
    model = fit_logit(binary_data, formula='class ~ x1 + x2')
    assert model._get_significance_stars(0.0005) == '***'
    assert model._get_significance_stars(0.005) == '**'
    assert model._get_significance_stars(0.03) == '*'
    assert model._get_significance_stars(0.08) == '.'
    assert model._get_significance_stars(0.5) == ''



"""
Entropy and information theory tools.
"""

from .metrics import calculate_entropy, information_gain
from .visualization import plot_entropy

__all__ = [
    'calculate_entropy',
    'information_gain',
    'plot_entropy',
]


"""
ITM 4150: Advanced Business Analytics and Visualization
Python toolkit for course materials at Cedarville University.
"""

__version__ = "1.1.0"
__author__ = "Dr. John D. Delano"

# Import commonly used functions for convenient access
from cuanalytics.datasets.loaders import (
    load_movie_reviews_data,
    load_sms_spam_data,
    load_mushroom_data,
    load_iris_data,
    load_breast_cancer_data,
    load_real_estate_data,
)
from cuanalytics.entropy.metrics import calculate_entropy, information_gain
from cuanalytics.entropy.visualization import plot_entropy, plot_entropy_rectangles
from cuanalytics.preprocessing.split import split_data
from cuanalytics.preprocessing.scale import scale_data
from cuanalytics.trees.decision_tree import fit_tree, SimpleDecisionTree
from cuanalytics.lda.discriminant import fit_lda, LDAModel
from cuanalytics.svm import fit_svm, SVMModel
from cuanalytics.regression.linear import fit_lm, LinearRegressionModel
from cuanalytics.regression.logistic import fit_logit, LogisticRegressionModel
from cuanalytics.nn.neuralnet import fit_nn, NeuralNetModel
from cuanalytics.similarity import euclidean, manhattan, cosine, jaccard
from cuanalytics.metrics import ConfusionMatrix
from cuanalytics.knn import fit_knn_classifier, KNNClassifierModel, fit_knn_regressor, KNNRegressorModel
from cuanalytics.clustering import fit_kmeans, KMeansModel, fit_hierarchical, HierarchicalClusteringModel
from cuanalytics.text import (
    fit_text_vectorizer,
    TextVectorizer,
    fit_topic_lda,
    LatentDirichletAllocationModel,
)
from cuanalytics.model_selection import (
    cross_validate,
    grid_search_cv,
    plot_learning_curves,
    plot_roc,
    plot_lift,
    plot_profit,
    plot_cumulative_response,
)

# Define what gets imported with "from cuanalytics import *"
__all__ = [
    'load_movie_reviews_data',
    'load_sms_spam_data',
    'load_mushroom_data',
    'load_iris_data',
    'load_breast_cancer_data',
    'load_real_estate_data',
    'calculate_entropy',
    'information_gain',
    'plot_entropy',
    'plot_entropy_rectangles',
    'split_data',
    'scale_data',
    'fit_tree',
    'SimpleDecisionTree',
    'fit_lda',
    'LDAModel',
    'fit_svm',
    'SVMModel',
    'fit_lm',
    'LinearRegressionModel',
    'fit_logit',
    'LogisticRegressionModel',
    'fit_nn',
    'NeuralNetModel',
    'euclidean',
    'manhattan',
    'cosine',
    'jaccard',
    'ConfusionMatrix',
    'fit_knn_classifier',
    'KNNClassifierModel',
    'fit_knn_regressor',
    'KNNRegressorModel',
    'fit_kmeans',
    'KMeansModel',
    'fit_hierarchical',
    'HierarchicalClusteringModel',
    'fit_text_vectorizer',
    'TextVectorizer',
    'fit_topic_lda',
    'LatentDirichletAllocationModel',
    'cross_validate',
    'grid_search_cv',
    'plot_learning_curves',
    'plot_roc',
    'plot_lift',
    'plot_profit',
    'plot_cumulative_response',
]



import numpy as np
import pandas as pd
import pytest

from cuanalytics.preprocessing.scale import scale_data


def test_scale_data_standard_excludes_cols():
    df = pd.DataFrame({
        'x1': [1.0, 2.0, 3.0],
        'x2': [10.0, 20.0, 30.0],
        'y': [0, 1, 0],
    })
    scaled, scaler = scale_data(df, exclude_cols=['y'])
    assert 'y' in scaled.columns
    assert np.allclose(scaled['y'], df['y'])
    assert not np.allclose(scaled['x1'], df['x1'])
    assert scaler is not None


def test_scale_data_minmax():
    df = pd.DataFrame({
        'x1': [0.0, 5.0, 10.0],
        'y': [1, 0, 1],
    })
    scaled, _ = scale_data(df, method='minmax', exclude_cols=['y'])
    assert scaled['x1'].min() == 0.0
    assert scaled['x1'].max() == 1.0


def test_scale_data_with_existing_scaler():
    train = pd.DataFrame({'x1': [0.0, 1.0, 2.0], 'y': [0, 1, 0]})
    test = pd.DataFrame({'x1': [1.0, 3.0], 'y': [1, 0]})
    train_scaled, scaler = scale_data(train, exclude_cols=['y'])
    test_scaled, _ = scale_data(test, exclude_cols=['y'], scaler=scaler)
    assert train_scaled['x1'].mean() == pytest.approx(0.0, abs=1e-7)
    assert 'x1' in test_scaled.columns


def test_scale_data_invalid_method():
    df = pd.DataFrame({'x1': [1.0, 2.0], 'y': [0, 1]})
    with pytest.raises(ValueError, match="method must be"):
        scale_data(df, method='unknown', exclude_cols=['y'])


def test_scale_data_skip_binary_default():
    df = pd.DataFrame({
        'x1': [1.0, 2.0, 3.0],
        'flag': [0, 1, 0],
        'y': [0, 1, 0],
    })
    scaled, _ = scale_data(df, exclude_cols=['y'])
    assert np.allclose(scaled['flag'], df['flag'])



import numpy as np
import pandas as pd
import requests
from io import StringIO
import matplotlib.pyplot as plt

def calculate_entropy(data, target_col=None, split_col=None):
    """
    Calculate Shannon entropy.

    Parameters:
    -----------
    data : pd.Series | list | pd.DataFrame
        Series/list of labels, or a DataFrame when target_col is provided.
    target_col : str | None
        Column to calculate entropy for when data is a DataFrame.
    split_col : str | None
        When provided with a DataFrame, returns weighted entropy after splitting
        target_col by split_col.
    """
    if isinstance(data, pd.DataFrame):
        if target_col is None:
            raise ValueError("target_col is required when data is a DataFrame")
        if split_col is None:
            return calculate_entropy(data[target_col])
        n = len(data)
        weighted_entropy = 0.0
        for val in data[split_col].unique():
            child = data[data[split_col] == val][target_col]
            weight = len(child) / n
            weighted_entropy += weight * calculate_entropy(child)
        return weighted_entropy

    y = pd.Series(data)
    counts = y.value_counts()
    probabilities = counts / len(y)
    probabilities = probabilities[probabilities > 0]
    if len(probabilities) == 0:
        return 0.0
    # Handle p=0 case automatically (0*log(0) = 0 by convention)
    entropy = -np.sum(probabilities * np.log2(probabilities))
    return entropy

def information_gain(df, feature, target_col='class'):
    """Calculate information gain from splitting on a feature."""
    parent_entropy = calculate_entropy(df, target_col=target_col)
    weighted_child_entropy = calculate_entropy(
        df, target_col=target_col, split_col=feature
    )
    return parent_entropy - weighted_child_entropy



# tests/test_lda.py
"""
Tests for Linear Discriminant Analysis
"""

import pytest
import pandas as pd
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from cuanalytics.lda import fit_lda, LDAModel


@pytest.fixture
def multiclass_data():
    """Simple 3-class classification dataset"""
    np.random.seed(42)
    df = pd.DataFrame({
        'x1': np.concatenate([
            np.random.randn(30) + 3,
            np.random.randn(30),
            np.random.randn(30) - 3
        ]),
        'x2': np.concatenate([
            np.random.randn(30) + 3,
            np.random.randn(30),
            np.random.randn(30) - 3
        ]),
        'class': ['A'] * 30 + ['B'] * 30 + ['C'] * 30
    })
    return df


@pytest.fixture
def binary_data():
    """Binary classification dataset"""
    np.random.seed(42)
    df = pd.DataFrame({
        'x1': np.concatenate([np.random.randn(40) + 2, np.random.randn(40) - 2]),
        'x2': np.concatenate([np.random.randn(40) + 2, np.random.randn(40) - 2]),
        'class': ['A'] * 40 + ['B'] * 40
    })
    return df


@pytest.fixture(autouse=True)
def close_plots():
    """Automatically close all plots after each test"""
    yield
    plt.close('all')


class TestLDABasic:
    """Test basic LDA functionality"""
    
    def test_fit_lda_basic(self, multiclass_data):
        """Test basic LDA fitting"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        
        assert hasattr(lda, 'lda')
        assert hasattr(lda, 'feature_names')
        assert hasattr(lda, 'classes')
        assert len(lda.classes) == 3
    
    def test_fit_with_n_components(self, multiclass_data):
        """Test fitting with specified n_components"""
        lda = fit_lda(multiclass_data, formula='class ~ .', n_components=1)
        assert lda.n_components == 1
    
    def test_fit_with_solver(self, multiclass_data):
        """Test fitting with different solver"""
        lda = fit_lda(multiclass_data, formula='class ~ .', solver='eigen')
        assert lda.solver == 'eigen'
    
    def test_direct_class_instantiation(self, multiclass_data):
        """Test creating LDAModel directly"""
        lda = LDAModel(multiclass_data, formula='class ~ .')
        assert lda is not None

    def test_fit_with_formula(self, multiclass_data):
        """Test fitting with formula input"""
        lda = fit_lda(multiclass_data, formula='class ~ x1 + x2')
        assert lda.target == 'class'
        assert set(lda.feature_names) == {'x1', 'x2'}

    def test_get_metrics(self, multiclass_data):
        """Test getting classification metrics"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        metrics = lda.get_metrics()

        assert metrics['model_type'] == 'lda'
        assert 'score' in metrics
        assert 'accuracy' in metrics['score']
        assert 'confusion_matrix' in metrics['score']
        assert 'per_class' in metrics['score']
        assert 'kappa' in metrics['score']
        assert 0 <= metrics['score']['accuracy'] <= 1
        assert isinstance(metrics['score']['confusion_matrix'], np.ndarray)
        assert metrics['score']['confusion_matrix'].shape == (3, 3)
        assert metrics['classes'] == lda.classes
        assert 'feature_importance' in metrics


class TestLDAValidation:
    """Test input validation"""
    
    def test_non_numeric_features_error(self, multiclass_data):
        """Test error when features are non-numeric"""
        df = multiclass_data.copy()
        df['categorical'] = ['cat1', 'cat2', 'cat3'] * 30
        
        with pytest.raises(ValueError, match="must be numeric"):
            fit_lda(df, formula='class ~ I(categorical)')

    def test_categorical_feature_with_C_allows_fit(self, multiclass_data):
        """Test categorical features work when wrapped with C()"""
        df = multiclass_data.copy()
        df['categorical'] = ['cat1', 'cat2', 'cat3'] * 30
        lda = fit_lda(df, formula='class ~ x1 + C(categorical)')
        assert isinstance(lda, LDAModel)
    
    def test_unfitted_model_error(self):
        """Test error when using unfitted model"""
        lda = LDAModel.__new__(LDAModel)
        # Don't set lda.lda
        
        with pytest.raises(RuntimeError, match="not been fitted"):
            lda.predict(pd.DataFrame({'x1': [1], 'x2': [2]}))


class TestLDAPrediction:
    """Test prediction methods"""
    
    def test_predict(self, multiclass_data):
        """Test basic prediction"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        predictions = lda.predict(multiclass_data)
        
        assert len(predictions) == len(multiclass_data)
        assert all(pred in lda.classes for pred in predictions)
    
    def test_predict_without_target(self, multiclass_data):
        """Test prediction on data without target column"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        X_test = multiclass_data.drop('class', axis=1)
        
        predictions = lda.predict(X_test)
        assert len(predictions) == len(X_test)

    def test_predict_with_formula(self, multiclass_data):
        """Test prediction when fit with formula"""
        lda = fit_lda(multiclass_data, formula='class ~ x1 + x2')
        X_test = multiclass_data.drop('class', axis=1)

        predictions = lda.predict(X_test)
        assert len(predictions) == len(X_test)
    
    def test_predict_proba(self, multiclass_data):
        """Test probability prediction"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        probabilities = lda.predict_proba(multiclass_data)
        
        assert probabilities.shape == (len(multiclass_data), len(lda.classes))
        assert np.allclose(probabilities.sum(axis=1), 1.0)  # Probabilities sum to 1
    
    def test_predict_proba_without_target(self, multiclass_data):
        """Test probability prediction without target column"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        X_test = multiclass_data.drop('class', axis=1)
        
        probabilities = lda.predict_proba(X_test)
        assert probabilities.shape[0] == len(X_test)
    
    def test_score(self, multiclass_data):
        """Test scoring"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        report = lda.score(multiclass_data)
        
        assert 0 <= report['accuracy'] <= 1
        assert report['accuracy'] > 0.7  # Should fit training data reasonably


class TestLDATransform:
    """Test transformation and dimensionality reduction"""
    
    def test_transform(self, multiclass_data):
        """Test transforming to LDA space"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        transformed = lda.transform(multiclass_data)
        
        assert transformed.shape[0] == len(multiclass_data)
        assert transformed.shape[1] <= len(lda.classes) - 1
    
    def test_transform_without_target(self, multiclass_data):
        """Test transform on data without target column"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        X_test = multiclass_data.drop('class', axis=1)
        
        transformed = lda.transform(X_test)
        assert transformed.shape[0] == len(X_test)
    
    def test_get_feature_importance(self, multiclass_data):
        """Test getting feature importance"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        importance = lda.get_feature_importance()
        
        assert isinstance(importance, pd.DataFrame)
        assert 'feature' in importance.columns
        assert 'importance' in importance.columns
        assert len(importance) == len(lda.feature_names)


class TestLDAVisualization:
    """Test visualization methods"""
    
    def test_visualize_multiclass(self, multiclass_data, monkeypatch):
        """Test visualization with 3 classes"""
        monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .')
        lda.visualize()
    
    def test_visualize_binary(self, binary_data, monkeypatch):
        """Test visualization with 2 classes (1D projection)"""
        monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
        
        lda = fit_lda(binary_data, formula='class ~ .')
        lda.visualize()
    
    def test_visualize_custom_figsize(self, multiclass_data, monkeypatch):
        """Test visualization with custom figsize"""
        monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .')
        lda.visualize(figsize=(10, 6))
    
    def test_visualize_features(self, multiclass_data, monkeypatch):
        """Test feature-specific visualization"""
        monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .')
        lda.visualize_features('x1', 'x2')
    
    def test_visualize_features_invalid_features(self, multiclass_data):
        """Test error with invalid feature names"""
        lda = fit_lda(multiclass_data, formula='class ~ .')
        
        with pytest.raises(ValueError, match="Features must be from"):  # Changed from "original features"
            lda.visualize_features('invalid1', 'invalid2')
    
    def test_visualize_features_custom_figsize(self, multiclass_data, monkeypatch):
        """Test feature visualization with custom figsize"""
        monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .')
        lda.visualize_features('x1', 'x2', figsize=(12, 10))

    def test_visualize_features_formula_runs(self, multiclass_data, monkeypatch):
        """Test visualize_features runs with formula usage"""
        lda = fit_lda(multiclass_data, formula='class ~ x1 + x2')

        monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
        lda.visualize_features('x1', 'x2')


class TestLDASummary:
    """Test summary method"""
    
    def test_summary_multiclass(self, multiclass_data, monkeypatch):
        """Test summary with multiclass data"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .')
        result = lda.summary()
        
        # Summary is display-only, should return None
        assert result is None
    
    def test_summary_binary(self, binary_data, monkeypatch):
        """Test summary with binary classification"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        lda = fit_lda(binary_data, formula='class ~ .')
        result = lda.summary()
        
        assert result is None
    
    def test_summary_contents(self, multiclass_data, monkeypatch):
        """Test that summary runs without error"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .')
        
        # Just verify it runs without error
        lda.summary()
        
        # Verify model attributes are accessible
        assert len(lda.classes) == 3
        assert len(lda.feature_names) == 2
    
    def test_summary_no_return(self, multiclass_data, monkeypatch):
        """Test that summary returns None (display only)"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .')
        result = lda.summary()
        
        assert result is None


class TestLDAEdgeCases:
    """Test edge cases"""
    
    def test_two_classes(self, binary_data, monkeypatch):
        """Test with 2 classes (minimum for LDA)"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        lda = fit_lda(binary_data, formula='class ~ .')
        assert len(lda.classes) == 2
        
        # Binary LDA has 1 discriminant
        transformed = lda.transform(binary_data)
        assert transformed.shape[1] == 1
    
    def test_many_classes(self, monkeypatch):
        """Test with many classes"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        np.random.seed(42)
        df = pd.DataFrame({
            'x1': np.random.randn(150),
            'x2': np.random.randn(150),
            'class': ['A'] * 30 + ['B'] * 30 + ['C'] * 30 + ['D'] * 30 + ['E'] * 30
        })
        
        lda = fit_lda(df, formula='class ~ .')
        assert len(lda.classes) == 5
        
        # n_components = min(n_features, n_classes - 1)
        transformed = lda.transform(df)
        assert transformed.shape[1] == min(2, 4)  # min(2 features, 5-1 classes)
    
    def test_many_features(self, monkeypatch):
        """Test with many features"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        np.random.seed(42)
        df = pd.DataFrame({
            f'x{i}': np.concatenate([
                np.random.randn(20) + i,
                np.random.randn(20),
                np.random.randn(20) - i
            ]) for i in range(10)
        })
        df['class'] = ['A'] * 20 + ['B'] * 20 + ['C'] * 20
        
        lda = fit_lda(df, formula='class ~ .')
        assert len(lda.feature_names) == 10
    
    def test_perfect_separation(self, monkeypatch):
        """Test with perfectly separable classes"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        df = pd.DataFrame({
            'x1': [1, 2, 3, 10, 11, 12, 20, 21, 22],
            'x2': [1, 2, 3, 10, 11, 12, 20, 21, 22],
            'class': ['A', 'A', 'A', 'B', 'B', 'B', 'C', 'C', 'C']
        })
        
        lda = fit_lda(df, formula='class ~ .')
        accuracy = lda.score(df)['accuracy']
        
        assert accuracy == 1.0  # Perfect separation


class TestLDAFeatureImportance:
    """Test feature importance without coefficients"""
    
    def test_feature_importance_no_coef(self, multiclass_data, monkeypatch):
        """Test that feature importance works when coef_ exists"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .')
        
        # LDA should have coef_ after fitting
        assert hasattr(lda.lda, 'coef_')
        
        importance = lda.get_feature_importance()
        assert len(importance) == len(lda.feature_names)


class TestLDASolvers:
    """Test different solver options"""
    
    def test_svd_solver(self, multiclass_data, monkeypatch):
        """Test with SVD solver"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .', solver='svd')
        assert lda.solver == 'svd'
    
    def test_lsqr_solver(self, multiclass_data, monkeypatch):
        """Test with LSQR solver"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .', solver='lsqr')
        assert lda.solver == 'lsqr'
    
    def test_eigen_solver(self, multiclass_data, monkeypatch):
        """Test with eigen solver"""
        monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
        
        lda = fit_lda(multiclass_data, formula='class ~ .', solver='eigen')
        assert lda.solver == 'eigen'



import numpy as np
import pandas as pd
import pytest

from cuanalytics import fit_kmeans
from cuanalytics.clustering.kmeans import KMeansModel


@pytest.fixture
def cluster_data():
    np.random.seed(42)
    x1 = np.random.normal(0, 1, 60)
    x2 = np.random.normal(0, 1, 60)
    return pd.DataFrame({'x1': x1, 'x2': x2})


def test_fit_kmeans(cluster_data):
    model = fit_kmeans(cluster_data, formula='~ x1 + x2', n_clusters=3)
    assert isinstance(model, KMeansModel)


def test_predict_returns_series(cluster_data):
    model = fit_kmeans(cluster_data, formula='~ x1 + x2', n_clusters=3)
    preds = model.predict(cluster_data)
    assert isinstance(preds, pd.Series)
    assert len(preds) == len(cluster_data)


def test_score_returns_metrics(cluster_data):
    model = fit_kmeans(cluster_data, formula='~ x1 + x2', n_clusters=3)
    metrics = model.score(cluster_data)
    assert 'inertia' in metrics


def test_get_metrics(cluster_data):
    model = fit_kmeans(cluster_data, formula='~ x1 + x2', n_clusters=3)
    metrics = model.get_metrics()
    assert metrics['model_type'] == 'kmeans'
    assert all(isinstance(k, int) for k in metrics['cluster_counts'].keys())
    assert all(isinstance(v, int) for v in metrics['cluster_counts'].values())


def test_missing_formula_raises(cluster_data):
    with pytest.raises(ValueError, match="Must provide 'formula'"):
        fit_kmeans(cluster_data, formula=None)


def test_non_numeric_feature_raises(cluster_data):
    df = cluster_data.copy()
    df['cat'] = ['a', 'b', 'c'] * 20
    with pytest.raises(ValueError, match="All features must be numeric"):
        fit_kmeans(df, formula='~ x1 + cat', n_clusters=3)


def test_unfitted_predict_raises(cluster_data):
    model = KMeansModel.__new__(KMeansModel)
    with pytest.raises(RuntimeError, match="not been fitted"):
        model.predict(cluster_data)


def test_visualize_runs(cluster_data, monkeypatch):
    model = fit_kmeans(cluster_data, formula='~ x1 + x2', n_clusters=3)
    monkeypatch.setattr("matplotlib.pyplot.show", lambda: None)
    model.visualize()


def test_describe_clusters_returns_labeled_dataframe(cluster_data):
    model = fit_kmeans(cluster_data, formula='~ x1 + x2', n_clusters=3, random_state=42)
    described = model.describe_clusters(max_depth=2)
    assert isinstance(described, pd.DataFrame)
    assert len(described) == len(cluster_data)
    assert 'cluster' in described.columns
    assert 'cluster_rule' in described.columns
    assert described['cluster'].notna().all()
    assert described['cluster_rule'].notna().all()


def test_describe_clusters_invalid_criterion_raises(cluster_data):
    model = fit_kmeans(cluster_data, formula='~ x1 + x2', n_clusters=3, random_state=42)
    with pytest.raises(ValueError, match="criterion must be 'entropy' or 'gini'"):
        model.describe_clusters(criterion='bad_criterion')



# cuanalytics/clustering/hierarchical.py
"""
Hierarchical clustering.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.cluster.hierarchy import dendrogram, linkage
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import silhouette_score
from sklearn.decomposition import PCA


class HierarchicalClusteringModel:
    """
    Hierarchical clustering model.
    """

    def __init__(self, df, formula, n_clusters=3, linkage='ward'):
        self.original_df = df
        self.formula = formula
        self.n_clusters = n_clusters
        self.linkage = linkage
        self.model_spec = None

        if formula is None:
            raise ValueError("Must provide 'formula' for model specification")

        try:
            from formulaic import model_matrix
        except ImportError:
            raise ImportError(
                "Formula support requires the 'formulaic' library.\n"
                "Install it with: pip install formulaic"
            )

        rhs = formula
        if '~' in formula:
            rhs = formula.split('~', 1)[1].strip()

        df_rhs = df.copy()
        model_matrices = model_matrix(rhs, df_rhs, output='pandas')

        if hasattr(model_matrices, 'rhs'):
            X = model_matrices.rhs
        else:
            X = model_matrices

        self.model_spec = getattr(model_matrices, 'model_spec', None)

        if 'Intercept' in X.columns:
            X = X.drop('Intercept', axis=1)

        self.X = X
        self.feature_names = list(self.X.columns)

        used_vars = set()
        if self.model_spec is not None:
            rhs_spec = getattr(self.model_spec, 'rhs', self.model_spec)
            used_vars = {var for var in getattr(rhs_spec, 'variables', set())
                         if var in df_rhs.columns}
        if not used_vars:
            used_vars = set(df_rhs.columns)
        non_numeric = [col for col in used_vars
                       if not pd.api.types.is_numeric_dtype(df_rhs[col])]
        if non_numeric:
            raise ValueError("All features must be numeric. Encode categorical variables first.")

        self._fit()
        self._print_fit_summary()

    def _fit(self):
        self.model = AgglomerativeClustering(n_clusters=self.n_clusters, linkage=self.linkage)
        self.labels_ = self.model.fit_predict(self.X)

    def _check_fitted(self):
        if not hasattr(self, 'model') or self.model is None:
            raise RuntimeError(
                "Model has not been fitted yet. "
                "Create model with: model = fit_hierarchical(df, formula='x1 + x2')"
            )

    def _print_fit_summary(self):
        print("\nHierarchical clustering fitted successfully!")
        print(f"  Clusters: {self.n_clusters}")
        print(f"  Features: {len(self.feature_names)}")
        print(f"  Samples: {len(self.X)}")
        print(f"  Linkage: {self.linkage}")

    def _transform_data_with_formula(self, df):
        if getattr(self, 'model_spec', None) is None:
            raise RuntimeError("Formula metadata missing; cannot transform new data.")

        rhs_spec = getattr(self.model_spec, 'rhs', self.model_spec)
        model_matrices = rhs_spec.get_model_matrix(df, output='pandas')

        if hasattr(model_matrices, 'rhs'):
            X_new = model_matrices.rhs
        else:
            X_new = model_matrices

        if 'Intercept' in X_new.columns:
            X_new = X_new.drop('Intercept', axis=1)

        return X_new

    def predict(self, df=None):
        self._check_fitted()
        if df is not None:
            raise ValueError("Hierarchical clustering does not support predicting new data.")
        return pd.Series(self.labels_, index=self.X.index, name='cluster')

    def score(self, df=None):
        self._check_fitted()
        X = self.X if df is None else self._transform_data_with_formula(df)
        labels = self.labels_ if df is None else AgglomerativeClustering(
            n_clusters=self.n_clusters, linkage=self.linkage
        ).fit_predict(X)
        metrics = {}
        if self.n_clusters > 1 and len(X) > self.n_clusters:
            metrics['silhouette'] = float(silhouette_score(X, labels))
        return metrics

    def get_metrics(self):
        self._check_fitted()
        metrics = {
            'model_type': 'hierarchical',
            'n_clusters': self.n_clusters,
            'n_samples': len(self.X),
            'n_features': len(self.feature_names),
            'feature_names': self.feature_names,
            'linkage': self.linkage,
            'cluster_counts': {
                int(cluster): int(count)
                for cluster, count in pd.Series(self.labels_).value_counts().sort_index().items()
            },
        }
        if self.n_clusters > 1 and len(self.X) > self.n_clusters:
            metrics['silhouette'] = float(silhouette_score(self.X, self.labels_))
        return metrics

    def visualize(
        self,
        figsize=None,
        leaf_rotation=90,
        leaf_font_size=8,
        cutoff=None,
        truncate_mode='lastp',
    ):
        self._check_fitted()
        Z = linkage(self.X, method=self.linkage)
        n_samples = len(self.X)
        valid_truncate_modes = {'lastp', 'level', None}
        if truncate_mode not in valid_truncate_modes:
            raise ValueError("truncate_mode must be one of: 'lastp', 'level', or None.")
        if cutoff is not None:
            if not isinstance(cutoff, int):
                raise ValueError("cutoff must be an integer.")
            min_cutoff = 0 if truncate_mode == 'level' else 2
            if cutoff < min_cutoff:
                raise ValueError(f"cutoff must be >= {min_cutoff} for truncate_mode='{truncate_mode}'.")
            if cutoff > n_samples:
                raise ValueError("cutoff cannot exceed the number of samples.")

        if figsize is None:
            width = max(10, min(40, n_samples * 0.18 if cutoff is None else cutoff * 0.6))
            figsize = (width, 6)

        fig, ax = plt.subplots(figsize=figsize)
        dendrogram_kwargs = {
            'ax': ax,
            'leaf_rotation': leaf_rotation,
            'leaf_font_size': leaf_font_size,
        }
        if cutoff is None:
            dendrogram_kwargs['labels'] = self.X.index.astype(str).tolist()
            title = 'Hierarchical Clustering Dendrogram'
        else:
            mode = 'lastp' if truncate_mode is None else truncate_mode
            if mode == 'level' and cutoff == 0:
                # Show only the root split as two top-level buckets.
                dendrogram_kwargs['truncate_mode'] = 'lastp'
                dendrogram_kwargs['p'] = 2
                dendrogram_kwargs['show_contracted'] = True
                title = 'Hierarchical Clustering Dendrogram (level=0, root split)'
            else:
                dendrogram_kwargs['truncate_mode'] = mode
                dendrogram_kwargs['p'] = cutoff
                dendrogram_kwargs['show_contracted'] = True
                title = f'Hierarchical Clustering Dendrogram ({mode}, cutoff={cutoff})'

        dendrogram(Z, **dendrogram_kwargs)
        ax.set_title(title)
        ax.set_xlabel('Sample Index')
        ax.set_ylabel('Distance')
        ax.grid(True, axis='y', alpha=0.3)
        plt.tight_layout()
        plt.show()

    def visualize_all_features(self, figsize=(8, 6)):
        self._check_fitted()
        if len(self.feature_names) > 2:
            pca = PCA(n_components=2)
            X_plot = pca.fit_transform(self.X)
            xlabel = 'PC1'
            ylabel = 'PC2'
        else:
            X_plot = self.X.values
            xlabel, ylabel = self.feature_names

        unique_labels = np.sort(np.unique(self.labels_))
        base_colors = [
            '#ff7f00', '#377eb8', '#4daf4a', '#984ea3', '#b8860b',
            '#a65628', '#f781bf', '#999999', '#66c2a5', '#1b9e77',
        ]
        marker_cycle = ['o', 's', '^', 'D', 'P', 'X', 'v', '<', '>', '*']
        colors = {label: base_colors[i % len(base_colors)] for i, label in enumerate(unique_labels)}
        markers = {label: marker_cycle[i % len(marker_cycle)] for i, label in enumerate(unique_labels)}

        fig, ax = plt.subplots(figsize=figsize)
        for label in unique_labels:
            mask = self.labels_ == label
            ax.scatter(
                X_plot[mask, 0],
                X_plot[mask, 1],
                color=colors[label],
                marker=markers[label],
                alpha=0.9,
                edgecolors='white',
                linewidth=0.7,
                label=f'Cluster {int(label)}',
            )
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title('Hierarchical Clusters (All Features)')
        ax.grid(True, alpha=0.3)
        plt.tight_layout()
        plt.show()

    def summary(self):
        self._check_fitted()
        metrics = self.get_metrics()
        print("\n" + "=" * 70)
        print("HIERARCHICAL CLUSTERING SUMMARY")
        print("=" * 70)
        print(f"Clusters: {self.n_clusters}")
        print(f"Features: {len(self.feature_names)}")
        print(f"Samples: {len(self.X)}")
        print(f"Linkage: {self.linkage}")
        if 'silhouette' in metrics:
            print(f"Silhouette: {metrics['silhouette']:.4f}")
        print("\nCluster Counts:")
        print(metrics['cluster_counts'])
        print("\n" + "=" * 70)


def fit_hierarchical(df, formula, n_clusters=3, linkage='ward'):
    if formula is None:
        raise ValueError("Must provide 'formula' for model specification")
    return HierarchicalClusteringModel(
        df,
        formula=formula,
        n_clusters=n_clusters,
        linkage=linkage,
    )



"""
Cross-validation utilities for cuanalytics models.
"""

from __future__ import annotations

import warnings
import numpy as np
import pandas as pd
from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.metrics import (
    accuracy_score,
    auc,
    cohen_kappa_score,
    f1_score,
    mean_absolute_error,
    mean_squared_error,
    roc_curve,
    r2_score,
)


def _parse_target(formula: str) -> str:
    if formula is None or "~" not in formula:
        raise ValueError("Formula must include a target (e.g., 'y ~ x1 + x2').")
    lhs, _rhs = formula.split("~", 1)
    target = lhs.strip()
    if not target:
        raise ValueError("Formula must include a target (e.g., 'y ~ x1 + x2').")
    return target


def _infer_task_type(y: pd.Series) -> str:
    if not pd.api.types.is_numeric_dtype(y):
        return "classification"
    if y.nunique() <= 10:
        return "classification"
    return "regression"


def _default_scoring(task: str) -> dict:
    if task == "classification":
        return {
            "accuracy": lambda y_true, y_pred: accuracy_score(y_true, y_pred),
            "kappa": lambda y_true, y_pred: cohen_kappa_score(y_true, y_pred),
            "macro_f1": lambda y_true, y_pred: f1_score(y_true, y_pred, average="macro"),
        }
    return {
        "r2": lambda y_true, y_pred: r2_score(y_true, y_pred),
        "rmse": lambda y_true, y_pred: np.sqrt(mean_squared_error(y_true, y_pred)),
        "mae": lambda y_true, y_pred: mean_absolute_error(y_true, y_pred),
    }


def cross_validate(
    fit_fn,
    df,
    formula,
    k=5,
    shuffle=True,
    random_state=None,
    stratify_on="auto",
    task="auto",
    scoring=None,
    return_models=False,
    return_oof=False,
    verbose=False,
    **model_kwargs,
):
    """
    Cross-validate a cuanalytics supervised model.

    Parameters:
    -----------
    fit_fn : callable
        Model fitting function (e.g., fit_lm, fit_logit).
    df : pd.DataFrame
        DataFrame containing features and target.
    formula : str
        R-style formula (e.g., 'y ~ x1 + x2').
    k : int
        Number of folds.
    shuffle : bool
        Whether to shuffle data before splitting.
    random_state : int | None
        Random seed for reproducibility (used if shuffle=True).
    stratify_on : str | "auto" | None
        Column to stratify on (classification only).
        - "auto": uses target column for classification tasks
        - None: no stratification
    task : str
        "auto", "classification", or "regression".
    scoring : dict[str, callable] | None
        Custom scoring functions. Each callable takes (y_true, y_pred).
    return_models : bool
        If True, returns list of fitted fold models.
    return_oof : bool
        If True, returns out-of-fold predictions as a pandas Series.
    verbose : bool
        If False, suppresses model fit output during cross-validation.
    **model_kwargs
        Additional arguments passed to fit_fn.

    Returns:
    --------
    results : dict
        Dictionary with per-fold metrics and summary statistics.
    """
    if k < 2:
        raise ValueError("k must be at least 2 for cross-validation.")

    target = _parse_target(formula)
    if target not in df.columns:
        raise ValueError(f"Target '{target}' not found in data.")

    y = df[target]
    resolved_task = task
    if task == "auto":
        resolved_task = _infer_task_type(y)
    if resolved_task not in {"classification", "regression"}:
        raise ValueError("task must be 'auto', 'classification', or 'regression'.")

    if stratify_on == "auto":
        stratify_on = target if resolved_task == "classification" else None
    if stratify_on is not None and stratify_on not in df.columns:
        raise KeyError(f"Stratification column '{stratify_on}' not found in DataFrame")

    scoring_fns = scoring if scoring is not None else _default_scoring(resolved_task)
    if not isinstance(scoring_fns, dict) or not scoring_fns:
        raise ValueError("scoring must be a non-empty dict of name -> callable.")

    if resolved_task == "classification" and stratify_on is not None:
        splitter = StratifiedKFold(
            n_splits=k,
            shuffle=shuffle,
            random_state=random_state if shuffle else None,
        )
        split_iter = splitter.split(df, df[stratify_on])
    else:
        splitter = KFold(
            n_splits=k,
            shuffle=shuffle,
            random_state=random_state if shuffle else None,
        )
        split_iter = splitter.split(df)

    fold_metrics = []
    models = [] if return_models else None
    oof_predictions = pd.Series(index=df.index, dtype=object) if return_oof else None

    for fold_idx, (train_idx, test_idx) in enumerate(split_iter, start=1):
        train_df = df.iloc[train_idx]
        test_df = df.iloc[test_idx]

        if verbose:
            model = fit_fn(train_df, formula=formula, **model_kwargs)
        else:
            import contextlib
            import io
            with contextlib.redirect_stdout(io.StringIO()):
                model = fit_fn(train_df, formula=formula, **model_kwargs)

        y_true = test_df[target]
        y_pred = model.predict(test_df)

        fold_result = {"fold": fold_idx}
        for name, fn in scoring_fns.items():
            fold_result[name] = fn(y_true, y_pred)
        fold_metrics.append(fold_result)

        if return_models:
            models.append(model)

        if return_oof:
            oof_predictions.iloc[test_idx] = y_pred

    metrics_df = pd.DataFrame(fold_metrics).set_index("fold")
    summary = {
        "mean": metrics_df.mean().to_dict(),
        "std": metrics_df.std(ddof=1).to_dict(),
    }

    results = {
        "task": resolved_task,
        "k": k,
        "stratify_on": stratify_on,
        "folds": fold_metrics,
        "summary": summary,
    }

    if return_models:
        results["models"] = models
    if return_oof:
        results["oof_predictions"] = oof_predictions

    return results


def grid_search_cv(
    fit_fn,
    df,
    formula,
    param_grid,
    k=5,
    shuffle=True,
    random_state=None,
    stratify_on="auto",
    task="auto",
    scoring=None,
    refit=None,
    verbose=False,
    **model_kwargs,
):
    """
    Grid-search hyperparameters using cross-validation and refit the best model.

    Parameters:
    -----------
    fit_fn : callable
        Model fitting function (e.g., fit_lm, fit_logit).
    df : pd.DataFrame
        DataFrame containing features and target.
    formula : str
        R-style formula (e.g., 'y ~ x1 + x2').
    param_grid : dict[str, list]
        Dictionary of hyperparameter names to candidate values.
    k : int
        Number of folds.
    shuffle : bool
        Whether to shuffle data before splitting.
    random_state : int | None
        Random seed for reproducibility (used if shuffle=True).
    stratify_on : str | "auto" | None
        Column to stratify on (classification only).
        - "auto": uses target column for classification tasks
        - None: no stratification
    task : str
        "auto", "classification", or "regression".
    scoring : dict[str, callable] | None
        Custom scoring functions. Each callable takes (y_true, y_pred).
    refit : str | None
        Metric name to choose the best params. Defaults to
        "accuracy" for classification and "rmse" for regression.
    verbose : bool
        If False, suppresses model fit output during CV runs.
    **model_kwargs
        Additional arguments passed to fit_fn.

    Returns:
    --------
    results : dict
        Dictionary with CV results and the refit best model.
    """
    import itertools

    if k < 2:
        raise ValueError("k must be at least 2 for cross-validation.")

    if not isinstance(param_grid, dict) or not param_grid:
        raise ValueError("param_grid must be a non-empty dict of name -> list of values.")
    for name, values in param_grid.items():
        if not isinstance(values, (list, tuple)) or len(values) == 0:
            raise ValueError(f"param_grid values for '{name}' must be a non-empty list.")

    target = _parse_target(formula)
    if target not in df.columns:
        raise ValueError(f"Target '{target}' not found in data.")

    y = df[target]
    resolved_task = task
    if task == "auto":
        resolved_task = _infer_task_type(y)
    if resolved_task not in {"classification", "regression"}:
        raise ValueError("task must be 'auto', 'classification', or 'regression'.")

    if stratify_on == "auto":
        stratify_on = target if resolved_task == "classification" else None
    if stratify_on is not None and stratify_on not in df.columns:
        raise KeyError(f"Stratification column '{stratify_on}' not found in DataFrame")

    scoring_fns = scoring if scoring is not None else _default_scoring(resolved_task)
    if not isinstance(scoring_fns, dict) or not scoring_fns:
        raise ValueError("scoring must be a non-empty dict of name -> callable.")

    if refit is None:
        refit = "accuracy" if resolved_task == "classification" else "rmse"
    if refit not in scoring_fns:
        raise ValueError(f"refit '{refit}' not found in scoring functions.")

    if resolved_task == "classification" and stratify_on is not None:
        splitter = StratifiedKFold(
            n_splits=k,
            shuffle=shuffle,
            random_state=random_state if shuffle else None,
        )
        split_iter = list(splitter.split(df, df[stratify_on]))
    else:
        splitter = KFold(
            n_splits=k,
            shuffle=shuffle,
            random_state=random_state if shuffle else None,
        )
        split_iter = list(splitter.split(df))

    param_names = list(param_grid.keys())
    param_values = [param_grid[name] for name in param_names]
    param_combinations = [
        dict(zip(param_names, values)) for values in itertools.product(*param_values)
    ]

    cv_results = []
    for params in param_combinations:
        fold_metrics = []
        for fold_idx, (train_idx, test_idx) in enumerate(split_iter, start=1):
            train_df = df.iloc[train_idx]
            test_df = df.iloc[test_idx]

            if verbose:
                model = fit_fn(train_df, formula=formula, **model_kwargs, **params)
            else:
                import contextlib
                import io
                with contextlib.redirect_stdout(io.StringIO()):
                    model = fit_fn(train_df, formula=formula, **model_kwargs, **params)

            y_true = test_df[target]
            y_pred = model.predict(test_df)

            fold_result = {"fold": fold_idx}
            for name, fn in scoring_fns.items():
                fold_result[name] = fn(y_true, y_pred)
            fold_metrics.append(fold_result)

        metrics_df = pd.DataFrame(fold_metrics).set_index("fold")
        summary = {
            "mean": metrics_df.mean().to_dict(),
            "std": metrics_df.std(ddof=1).to_dict(),
        }

        cv_results.append(
            {
                "params": params,
                "folds": fold_metrics,
                "mean": summary["mean"],
                "std": summary["std"],
            }
        )

    lower_is_better = {"rmse", "mae", "mse"}
    best_idx = None
    best_score = None
    for idx, result in enumerate(cv_results):
        score = result["mean"][refit]
        if best_idx is None:
            best_idx = idx
            best_score = score
            continue
        if refit in lower_is_better:
            if score < best_score:
                best_idx = idx
                best_score = score
        else:
            if score > best_score:
                best_idx = idx
                best_score = score

    best_params = cv_results[best_idx]["params"]
    if verbose:
        best_model = fit_fn(df, formula=formula, **model_kwargs, **best_params)
    else:
        import contextlib
        import io
        with contextlib.redirect_stdout(io.StringIO()):
            best_model = fit_fn(df, formula=formula, **model_kwargs, **best_params)

    return {
        "task": resolved_task,
        "k": k,
        "stratify_on": stratify_on,
        "refit": refit,
        "best_params": best_params,
        "best_score": best_score,
        "best_model": best_model,
        "cv_results": cv_results,
    }


def plot_learning_curves(
    fit_fns,
    df,
    formula,
    train_sizes=None,
    k=5,
    shuffle=True,
    random_state=None,
    stratify_on="auto",
    task="auto",
    scoring=None,
    metric=None,
    verbose=False,
    figsize=(10, 6),
    title=None,
    **model_kwargs,
):
    """
    Plot learning curves (validation performance vs training size) for one or more models.

    Parameters:
    -----------
    fit_fns : callable | list[callable]
        Model fitting function(s) (e.g., fit_lm, fit_logit).
    df : pd.DataFrame
        DataFrame containing features and target.
    formula : str
        R-style formula (e.g., 'y ~ x1 + x2').
    train_sizes : list[float|int] | None
        Fractions (0-1] or absolute sizes of training data to use per fold.
        If None, defaults to [0.1, 0.2, 0.4, 0.6, 0.8, 1.0].
    k : int
        Number of folds.
    shuffle : bool
        Whether to shuffle data before splitting.
    random_state : int | None
        Random seed for reproducibility (used if shuffle=True).
    stratify_on : str | "auto" | None
        Column to stratify on (classification only).
        - "auto": uses target column for classification tasks
        - None: no stratification
    task : str
        "auto", "classification", or "regression".
    scoring : dict[str, callable] | None
        Custom scoring functions. Each callable takes (y_true, y_pred).
    metric : str | None
        Metric name to plot (must be a key in scoring). Defaults to
        "accuracy" for classification and "rmse" for regression.
    verbose : bool
        If False, suppresses model fit output during learning curve runs.
    figsize : tuple
        Matplotlib figure size.
    title : str | None
        Plot title.
    **model_kwargs
        Additional arguments passed to fit_fn.

    Returns:
    --------
    results : dict
        Dictionary with learning curve results and matplotlib Axes.
    """
    import matplotlib.pyplot as plt

    if k < 2:
        raise ValueError("k must be at least 2 for learning curves.")

    target = _parse_target(formula)
    if target not in df.columns:
        raise ValueError(f"Target '{target}' not found in data.")

    y = df[target]
    resolved_task = task
    if task == "auto":
        resolved_task = _infer_task_type(y)
    if resolved_task not in {"classification", "regression"}:
        raise ValueError("task must be 'auto', 'classification', or 'regression'.")

    if stratify_on == "auto":
        stratify_on = target if resolved_task == "classification" else None
    if stratify_on is not None and stratify_on not in df.columns:
        raise KeyError(f"Stratification column '{stratify_on}' not found in DataFrame")

    scoring_fns = scoring if scoring is not None else _default_scoring(resolved_task)
    if not isinstance(scoring_fns, dict) or not scoring_fns:
        raise ValueError("scoring must be a non-empty dict of name -> callable.")

    if metric is None:
        metric = "accuracy" if resolved_task == "classification" else "rmse"
    if metric not in scoring_fns:
        raise ValueError(f"metric '{metric}' not found in scoring functions.")

    if train_sizes is None:
        train_sizes = [0.1, 0.2, 0.4, 0.6, 0.8, 1.0]

    if not isinstance(fit_fns, (list, tuple)):
        fit_fns = [fit_fns]
    if not fit_fns:
        raise ValueError("fit_fns must include at least one model function.")
    show_band = len(fit_fns) == 1

    if resolved_task == "classification" and stratify_on is not None:
        splitter = StratifiedKFold(
            n_splits=k,
            shuffle=shuffle,
            random_state=random_state if shuffle else None,
        )
        split_iter = list(splitter.split(df, df[stratify_on]))
    else:
        splitter = KFold(
            n_splits=k,
            shuffle=shuffle,
            random_state=random_state if shuffle else None,
        )
        split_iter = list(splitter.split(df))

    rng = np.random.default_rng(random_state) if shuffle else None
    base_train_len = len(split_iter[0][0]) if split_iter else 0
    plot_x = []
    for size in train_sizes:
        if isinstance(size, float):
            if size <= 0 or size > 1:
                raise ValueError("train_sizes fractions must be in (0, 1].")
            n_train = max(1, int(round(size * base_train_len)))
        else:
            n_train = int(size)
            if n_train <= 0:
                raise ValueError("train_sizes must be positive.")
        n_train = min(n_train, base_train_len)
        plot_x.append(n_train)

    results = {
        "task": resolved_task,
        "train_sizes": train_sizes,
        "train_sizes_abs": plot_x,
        "metric": metric,
        "models": {},
    }

    fig, ax = plt.subplots(figsize=figsize)
    band_label = None

    for fit_fn in fit_fns:
        model_name = getattr(fit_fn, "__name__", "model")
        val_means = []
        val_stds = []

        for size in train_sizes:
            fold_scores = []
            for train_idx, test_idx in split_iter:
                train_idx = np.array(train_idx)
                if shuffle and rng is not None:
                    rng.shuffle(train_idx)

                if isinstance(size, float):
                    if size <= 0 or size > 1:
                        raise ValueError("train_sizes fractions must be in (0, 1].")
                    n_train = max(1, int(round(size * len(train_idx))))
                else:
                    n_train = int(size)
                    if n_train <= 0:
                        raise ValueError("train_sizes must be positive.")
                n_train = min(n_train, len(train_idx))

                subset_idx = train_idx[:n_train]
                train_df = df.iloc[subset_idx]
                test_df = df.iloc[test_idx]

                if verbose:
                    model = fit_fn(train_df, formula=formula, **model_kwargs)
                else:
                    import contextlib
                    import io
                    with contextlib.redirect_stdout(io.StringIO()):
                        model = fit_fn(train_df, formula=formula, **model_kwargs)

                y_true = test_df[target]
                y_pred = model.predict(test_df)
                fold_scores.append(scoring_fns[metric](y_true, y_pred))

            val_means.append(float(np.mean(fold_scores)))
            val_stds.append(float(np.std(fold_scores, ddof=1)) if len(fold_scores) > 1 else 0.0)

        results["models"][model_name] = {
            "val_mean": val_means,
            "val_std": val_stds,
        }

        ax.plot(plot_x, val_means, marker="o", label=model_name)
        if show_band:
            band_label = band_label or "±1 std (validation)"
            ax.fill_between(
                plot_x,
                np.array(val_means) - np.array(val_stds),
                np.array(val_means) + np.array(val_stds),
                alpha=0.15,
                label=band_label,
            )
            band_label = None

    ax.set_xlabel("Training Set Size")
    ax.set_ylabel(metric)
    ax.set_title(title or f"Learning Curve ({metric})")
    ax.grid(True, alpha=0.3)
    ax.legend(loc="lower right")

    results["ax"] = ax
    return results


def _as_model_list(models):
    if isinstance(models, (list, tuple)):
        model_list = list(models)
    else:
        model_list = [models]
    if not model_list:
        raise ValueError("models must include at least one fitted classifier.")
    return model_list


def _resolve_model_names(models, model_names=None):
    if model_names is None:
        names = []
        seen = {}
        for model in models:
            base = type(model).__name__
            seen[base] = seen.get(base, 0) + 1
            suffix = f" #{seen[base]}" if seen[base] > 1 else ""
            names.append(f"{base}{suffix}")
        return names
    if len(model_names) != len(models):
        raise ValueError("model_names length must match the number of models.")
    return list(model_names)


def _resolve_positive_class(y_true, positive_class=None):
    classes = list(pd.Series(y_true).dropna().unique())
    if len(classes) != 2:
        raise ValueError("ROC/Lift/Profit currently require binary classification targets.")
    if positive_class is not None:
        if positive_class not in classes:
            raise ValueError(f"positive_class '{positive_class}' not found in y_true classes {classes}.")
        return positive_class
    class_set = set(classes)
    if class_set == {0, 1}:
        return 1
    if class_set == {"0", "1"}:
        return "1"
    return sorted(classes)[-1]


def _resolve_sample_weight(sample_weight, test_df):
    if sample_weight is None:
        return np.ones(len(test_df), dtype=float)
    if isinstance(sample_weight, pd.Series):
        w = sample_weight.reindex(test_df.index).to_numpy(dtype=float)
    else:
        w = np.asarray(sample_weight, dtype=float)
    if len(w) != len(test_df):
        raise ValueError("sample_weight length must match test_df length.")
    return w


def _get_model_classes(model):
    if hasattr(model, "target_encoder") and hasattr(model.target_encoder, "classes_"):
        return list(model.target_encoder.classes_)
    if hasattr(model, "model") and hasattr(model.model, "classes_"):
        return list(model.model.classes_)
    if hasattr(model, "tree") and hasattr(model.tree, "classes_"):
        return list(model.tree.classes_)
    if hasattr(model, "lda") and hasattr(model.lda, "classes_"):
        return list(model.lda.classes_)
    if hasattr(model, "classes"):
        return list(model.classes)
    return None


def _extract_positive_scores(model, test_df, positive_class):
    if not hasattr(model, "predict_proba"):
        raise ValueError(
            f"{type(model).__name__} does not implement predict_proba(df). "
            "ROC/Lift/Profit require probability scores."
        )

    proba = model.predict_proba(test_df)
    proba = np.asarray(proba, dtype=float)
    if proba.ndim == 1:
        return proba
    if proba.ndim != 2:
        raise ValueError(f"{type(model).__name__}.predict_proba returned unexpected shape: {proba.shape}")
    if proba.shape[1] < 2:
        raise ValueError(f"{type(model).__name__}.predict_proba must return at least 2 columns for binary tasks.")

    classes = _get_model_classes(model)
    if classes is not None:
        classes = list(classes)
        if positive_class in classes:
            return proba[:, classes.index(positive_class)]

    if proba.shape[1] == 2:
        warnings.warn(
            f"Could not resolve classes for {type(model).__name__}; using predict_proba[:, 1] for positive class.",
            RuntimeWarning,
            stacklevel=2,
        )
        return proba[:, 1]
    raise ValueError(
        f"Could not map positive_class '{positive_class}' to predict_proba columns for {type(model).__name__}."
    )


class CurveAnalysisResult:
    """
    Holds model scores and cached curve calculations for ROC/Lift/Profit analysis.
    """

    def __init__(
        self,
        models,
        model_names,
        test_df,
        y_true,
        positive_class,
        sample_weight,
        profit_config,
        scores_by_model,
        target,
    ):
        self.models = models
        self.model_names = model_names
        self.test_df = test_df
        self.y_true = pd.Series(y_true, index=test_df.index)
        self.target = target
        self.positive_class = positive_class
        self.sample_weight = sample_weight
        self.profit_config = profit_config
        self.scores_by_model = scores_by_model
        self._curve_cache = {}

    def _calculate_curve(self, curve):
        curve = str(curve).lower()
        if curve in self._curve_cache:
            return self._curve_cache[curve]
        if curve not in {"roc", "lift", "profit", "cumulative_response"}:
            raise ValueError("curve must be one of {'roc', 'lift', 'profit', 'cumulative_response'}.")

        y_bin = (self.y_true.values == self.positive_class).astype(int)
        w = self.sample_weight
        curve_result = {}

        for model_name in self.model_names:
            score = self.scores_by_model[model_name]

            if curve == "roc":
                fpr, tpr, thresholds = roc_curve(y_bin, score, sample_weight=w)
                model_df = pd.DataFrame({
                    "threshold": thresholds,
                    "fpr": fpr,
                    "tpr": tpr,
                })
                finite_mask = np.isfinite(thresholds)
                if np.any(finite_mask):
                    dist = np.hypot(fpr[finite_mask], 1 - tpr[finite_mask])
                    best_local_idx = int(np.argmin(dist))
                    best_idx = np.where(finite_mask)[0][best_local_idx]
                else:
                    dist = np.hypot(fpr, 1 - tpr)
                    best_idx = int(np.argmin(dist))
                curve_result[model_name] = {
                    "data": model_df,
                    "auc": float(auc(fpr, tpr)),
                    "best_cutoff": float(thresholds[best_idx]),
                    "best_fpr": float(fpr[best_idx]),
                    "best_tpr": float(tpr[best_idx]),
                    "best_distance": float(np.hypot(fpr[best_idx], 1 - tpr[best_idx])),
                }
                continue

            if curve in {"lift", "cumulative_response"}:
                order = np.argsort(-score, kind="mergesort")
                y_sorted = y_bin[order]
                w_sorted = w[order]
                score_sorted = score[order]
                cum_w = np.cumsum(w_sorted)
                cum_pos = np.cumsum(w_sorted * y_sorted)
                total_w = float(cum_w[-1]) if len(cum_w) else 0.0
                total_pos = float(cum_pos[-1]) if len(cum_pos) else 0.0

                pop_frac = cum_w / total_w if total_w else np.zeros_like(cum_w)
                capture = cum_pos / total_pos if total_pos else np.zeros_like(cum_pos)
                with np.errstate(divide="ignore", invalid="ignore"):
                    lift = np.where(pop_frac > 0, capture / pop_frac, 1.0)

                model_df = pd.DataFrame({
                    "threshold": score_sorted,
                    "population_frac": pop_frac,
                    "capture_rate": capture,
                    "lift": lift,
                })
                model_df = pd.concat(
                    [pd.DataFrame([{
                        "threshold": 1.0,
                        "population_frac": 0.0,
                        "capture_rate": 0.0,
                        "lift": 1.0,
                    }]), model_df],
                    ignore_index=True,
                )
                top_decile = model_df.loc[model_df["population_frac"] >= 0.10, "lift"]
                top_decile_lift = float(top_decile.iloc[0]) if not top_decile.empty else float(model_df["lift"].iloc[-1])
                curve_result[model_name] = {
                    "data": model_df,
                    "top_decile_lift": top_decile_lift,
                }
                continue

            order = np.argsort(-score, kind="mergesort")
            y_sorted = y_bin[order]
            w_sorted = w[order]
            score_sorted = score[order]

            total_w = float(np.sum(w_sorted))
            total_pos_w = float(np.sum(w_sorted * y_sorted))
            total_neg_w = float(np.sum(w_sorted * (1 - y_sorted)))

            tp_cum = np.cumsum(w_sorted * y_sorted)
            fp_cum = np.cumsum(w_sorted * (1 - y_sorted))
            fn_cum = total_pos_w - tp_cum
            tn_cum = total_neg_w - fp_cum
            pop_frac = tp_cum.copy()
            if total_w > 0:
                pop_frac = np.cumsum(w_sorted) / total_w
            else:
                pop_frac = np.zeros_like(tp_cum)

            # Baseline at threshold=1.0 means no one is targeted.
            tp = np.concatenate(([0.0], tp_cum))
            fp = np.concatenate(([0.0], fp_cum))
            fn = np.concatenate(([total_pos_w], fn_cum))
            tn = np.concatenate(([total_neg_w], tn_cum))
            thresholds = np.concatenate(([1.0], score_sorted))
            population_frac = np.concatenate(([0.0], pop_frac))

            conf = self.profit_config
            profit = (
                tp * conf["tp_value"]
                + fp * conf["fp_value"]
                + tn * conf["tn_value"]
                + fn * conf["fn_value"]
                + conf["fixed_value"]
            )

            model_df = pd.DataFrame({
                "threshold": thresholds,
                "population_frac": population_frac,
                "tp": tp,
                "fp": fp,
                "tn": tn,
                "fn": fn,
                "profit": profit,
            })
            best_idx = int(np.argmax(profit))
            curve_result[model_name] = {
                "data": model_df,
                "max_profit": float(profit[best_idx]),
                "max_profit_threshold": float(thresholds[best_idx]),
                "max_profit_population_frac": float(population_frac[best_idx]),
            }

        self._curve_cache[curve] = curve_result
        return curve_result

    def get_roc(self):
        return self._calculate_curve("roc")

    def get_lift(self):
        return self._calculate_curve("lift")

    def get_profit(self):
        return self._calculate_curve("profit")

    def get_cumulative_response(self):
        return self._calculate_curve("cumulative_response")

    def summary_table(self):
        roc = self.get_roc()
        lift = self.get_lift()
        profit = self.get_profit()
        rows = []
        for name in self.model_names:
            rows.append({
                "model": name,
                "auc": roc[name]["auc"],
                "top_decile_lift": lift[name]["top_decile_lift"],
                "max_profit": profit[name]["max_profit"],
                "max_profit_threshold": profit[name]["max_profit_threshold"],
            })
        return pd.DataFrame(rows).set_index("model")


def calculate_curve(
    models,
    test_df,
    positive_class=None,
    profit_config=None,
    sample_weight=None,
    model_names=None,
):
    """
    Build a curve-analysis result object for ROC/Lift/Profit comparisons.
    """
    models = _as_model_list(models)
    model_names = _resolve_model_names(models, model_names=model_names)

    first_model = models[0]
    target = getattr(first_model, "target", None)
    if target is None:
        raise ValueError("Could not infer target column from model. Expected model.target.")
    if target not in test_df.columns:
        raise ValueError(f"Target '{target}' not found in test_df.")
    for model in models[1:]:
        model_target = getattr(model, "target", None)
        if model_target != target:
            raise ValueError("All models must share the same target column.")

    y_true = test_df[target]
    positive_class = _resolve_positive_class(y_true, positive_class=positive_class)
    w = _resolve_sample_weight(sample_weight, test_df)

    default_profit = {
        "tp_value": 1.0,
        "fp_value": -1.0,
        "tn_value": 0.0,
        "fn_value": -1.0,
        "fixed_value": 0.0,
    }
    if profit_config is None:
        resolved_profit = default_profit
    else:
        resolved_profit = default_profit.copy()
        resolved_profit.update(profit_config)
    required = {"tp_value", "fp_value", "tn_value", "fn_value", "fixed_value"}
    missing = required - set(resolved_profit.keys())
    if missing:
        raise ValueError(f"profit_config missing required keys: {sorted(missing)}")

    scores_by_model = {}
    for model, name in zip(models, model_names):
        scores = _extract_positive_scores(model, test_df, positive_class)
        if len(scores) != len(test_df):
            raise ValueError(f"{name}.predict_proba returned {len(scores)} rows, expected {len(test_df)}.")
        scores_by_model[name] = np.asarray(scores, dtype=float)

    return CurveAnalysisResult(
        models=models,
        model_names=model_names,
        test_df=test_df,
        y_true=y_true,
        positive_class=positive_class,
        sample_weight=w,
        profit_config=resolved_profit,
        scores_by_model=scores_by_model,
        target=target,
    )


def plot_roc(
    models,
    test_df=None,
    positive_class=None,
    sample_weight=None,
    model_names=None,
    show_cutoffs=False,
    cutoff_step=0.1,
    cutoff_fontsize=8,
    show=True,
    ax=None,
    figsize=(8, 6),
    title="ROC Curve",
):
    """
    Plot ROC comparison and return ROC curve data.
    """
    import matplotlib.pyplot as plt

    if isinstance(models, CurveAnalysisResult):
        curve_result = models
    else:
        if test_df is None:
            raise ValueError("test_df is required when passing models to plot_roc.")
        curve_result = calculate_curve(
            models=models,
            test_df=test_df,
            positive_class=positive_class,
            sample_weight=sample_weight,
            model_names=model_names,
        )
    if ax is None:
        _fig, ax = plt.subplots(figsize=figsize)

    roc = curve_result.get_roc()
    for name in curve_result.model_names:
        data = roc[name]["data"]
        line, = ax.plot(data["fpr"], data["tpr"], label=f"{name} (AUC={roc[name]['auc']:.3f})")

        if show_cutoffs:
            if cutoff_step <= 0 or cutoff_step > 1:
                raise ValueError("cutoff_step must be in (0, 1].")

            finite_mask = np.isfinite(data["threshold"].to_numpy())
            finite_df = data.loc[finite_mask].copy()
            if not finite_df.empty:
                thresholds = finite_df["threshold"].to_numpy(dtype=float)
                fprs = finite_df["fpr"].to_numpy(dtype=float)
                tprs = finite_df["tpr"].to_numpy(dtype=float)

                desired = np.arange(0.0, 1.0 + 1e-9, cutoff_step)
                used_idx = set()
                for cutoff in desired:
                    idx = int(np.argmin(np.abs(thresholds - cutoff)))
                    if idx in used_idx:
                        continue
                    used_idx.add(idx)
                    x = float(fprs[idx])
                    y = float(tprs[idx])
                    ax.scatter([x], [y], s=14, color=line.get_color(), zorder=4)
                    ax.annotate(
                        f"{cutoff:.1f}",
                        (x, y),
                        textcoords="offset points",
                        xytext=(4, 4),
                        fontsize=cutoff_fontsize,
                        color=line.get_color(),
                        alpha=0.9,
                    )

            best_x = roc[name]["best_fpr"]
            best_y = roc[name]["best_tpr"]
            ax.scatter(
                [best_x],
                [best_y],
                marker="*",
                s=120,
                facecolor=line.get_color(),
                edgecolor="black",
                linewidth=0.7,
                zorder=5,
            )
            ax.annotate(
                f"best={roc[name]['best_cutoff']:.2f}",
                (best_x, best_y),
                textcoords="offset points",
                xytext=(6, -10),
                fontsize=max(7, cutoff_fontsize - 1),
                color=line.get_color(),
                alpha=0.95,
            )

    ax.plot([0, 1], [0, 1], linestyle="--", color="#6f665b", label="Random baseline")
    ax.set_xlabel("False Positive Rate")
    ax.set_ylabel("True Positive Rate")
    ax.set_title(title)
    ax.grid(True, alpha=0.3)
    ax.legend(loc="lower right")
    if show:
        plt.show()
    return {"ax": ax, "data": roc}


def plot_lift(
    models,
    test_df=None,
    positive_class=None,
    sample_weight=None,
    model_names=None,
    show=True,
    ax=None,
    figsize=(8, 6),
    title="Lift Curve",
):
    """
    Plot lift comparison and return lift curve data.
    """
    import matplotlib.pyplot as plt

    if isinstance(models, CurveAnalysisResult):
        curve_result = models
    else:
        if test_df is None:
            raise ValueError("test_df is required when passing models to plot_lift.")
        curve_result = calculate_curve(
            models=models,
            test_df=test_df,
            positive_class=positive_class,
            sample_weight=sample_weight,
            model_names=model_names,
        )
    if ax is None:
        _fig, ax = plt.subplots(figsize=figsize)

    lift = curve_result.get_lift()
    for name in curve_result.model_names:
        data = lift[name]["data"]
        ax.plot(data["population_frac"] * 100, data["lift"], label=f"{name} (Top 10%={lift[name]['top_decile_lift']:.2f}x)")

    ax.axhline(1.0, linestyle="--", color="#6f665b", label="Baseline")
    ax.set_xlabel("Population Contacted (%)")
    ax.set_ylabel("Lift")
    ax.set_title(title)
    ax.grid(True, alpha=0.3)
    ax.legend(loc="upper right")
    if show:
        plt.show()
    return {"ax": ax, "data": lift}


def plot_profit(
    models,
    test_df=None,
    positive_class=None,
    profit_config=None,
    sample_weight=None,
    model_names=None,
    show=True,
    ax=None,
    figsize=(8, 6),
    title="Profit Curve",
):
    """
    Plot profit vs targeted test population (%) and return profit curve data.
    """
    import matplotlib.pyplot as plt

    if isinstance(models, CurveAnalysisResult):
        curve_result = models
    else:
        if test_df is None:
            raise ValueError("test_df is required when passing models to plot_profit.")
        curve_result = calculate_curve(
            models=models,
            test_df=test_df,
            positive_class=positive_class,
            profit_config=profit_config,
            sample_weight=sample_weight,
            model_names=model_names,
        )
    if ax is None:
        _fig, ax = plt.subplots(figsize=figsize)

    profit = curve_result.get_profit()
    for name in curve_result.model_names:
        data = profit[name]["data"]
        ax.plot(data["population_frac"] * 100, data["profit"], label=name)
        max_x = profit[name].get("max_profit_population_frac")
        if max_x is None:
            max_x = float(data.loc[data["profit"].idxmax(), "population_frac"])
        ax.scatter(
            [max_x * 100],
            [profit[name]["max_profit"]],
            s=30,
            zorder=3,
        )

    ax.set_xlabel("Test Instances Targeted (%)")
    ax.set_ylabel("Profit")
    ax.set_title(title)
    ax.grid(True, alpha=0.3)
    ax.legend(loc="best")
    if show:
        plt.show()
    return {"ax": ax, "data": profit}


def plot_cumulative_response(
    models,
    test_df=None,
    positive_class=None,
    sample_weight=None,
    model_names=None,
    show=True,
    ax=None,
    figsize=(8, 6),
    title="Cumulative Response Curve",
):
    """
    Plot cumulative response (capture rate) vs population percentage.
    Returns the same underlying curve data used for lift.
    """
    import matplotlib.pyplot as plt

    if isinstance(models, CurveAnalysisResult):
        curve_result = models
    else:
        if test_df is None:
            raise ValueError("test_df is required when passing models to plot_cumulative_response.")
        curve_result = calculate_curve(
            models=models,
            test_df=test_df,
            positive_class=positive_class,
            sample_weight=sample_weight,
            model_names=model_names,
        )
    if ax is None:
        _fig, ax = plt.subplots(figsize=figsize)

    cumulative = curve_result.get_cumulative_response()
    for name in curve_result.model_names:
        data = cumulative[name]["data"]
        ax.plot(data["population_frac"] * 100, data["capture_rate"] * 100, label=name)

    ax.plot([0, 100], [0, 100], linestyle="--", color="#6f665b", label="Random baseline")
    ax.set_xlabel("Population Contacted (%)")
    ax.set_ylabel("Cumulative Response / Capture (%)")
    ax.set_title(title)
    ax.grid(True, alpha=0.3)
    ax.legend(loc="lower right")
    if show:
        plt.show()
    return {"ax": ax, "data": cumulative}



import numpy as np

from cuanalytics.similarity import euclidean, manhattan, cosine, jaccard


def test_euclidean():
    assert euclidean([0, 0], [3, 4]) == 5.0


def test_manhattan():
    assert manhattan([1, 2], [4, 6]) == 7.0


def test_cosine():
    assert np.isclose(cosine([1, 0], [0, 1]), 1.0)


def test_jaccard():
    assert np.isclose(jaccard([1, 0, 1], [1, 1, 0]), 2/3)



import pandas as pd
import requests
from io import StringIO
import io
import tarfile
import zipfile

def load_sms_spam_data():
    """
    Load the UCI SMS Spam Collection dataset.

    This dataset contains SMS messages labeled as either ham (legitimate)
    or spam (unsolicited commercial message).

    Returns:
    --------
    pd.DataFrame
        DataFrame with 5572 samples and 2 columns: 'label' and 'text'

    Examples:
    ---------
    >>> from cuanalytics.datasets import load_sms_spam_data
    >>> df = load_sms_spam_data()
    >>> print(df.shape)
    (5572, 2)
    >>> print(df['label'].value_counts())

    References:
    -----------
    https://archive.ics.uci.edu/dataset/228/sms+spam+collection
    """
    url = "https://archive.ics.uci.edu/ml/machine-learning-databases/00228/smsspamcollection.zip"

    try:
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        with zipfile.ZipFile(io.BytesIO(response.content)) as zip_file:
            with zip_file.open("SMSSpamCollection") as data_file:
                df = pd.read_csv(
                    data_file,
                    sep="\t",
                    names=["label", "text"],
                    encoding="utf-8",
                )
    except Exception as e:
        print(f"Error loading SMS spam data: {e}")
        print("Please ensure you have an internet connection.")
        raise

    return df

def load_movie_reviews_data():
    """
    Load the Cornell Movie Review Polarity dataset.

    This dataset contains 2,000 movie reviews labeled as either
    positive or negative sentiment.

    Returns:
    --------
    pd.DataFrame
        DataFrame with 2000 samples and 2 columns: 'label' and 'text'

    Examples:
    ---------
    >>> from cuanalytics.datasets import load_movie_reviews_data
    >>> df = load_movie_reviews_data()
    >>> print(df.shape)
    (2000, 2)
    >>> print(df['label'].value_counts())

    References:
    -----------
    https://www.cs.cornell.edu/people/pabo/movie-review-data/
    """
    url = "https://www.cs.cornell.edu/people/pabo/movie-review-data/review_polarity.tar.gz"

    try:
        response = requests.get(url, timeout=20)
        response.raise_for_status()

        records = []
        with tarfile.open(fileobj=io.BytesIO(response.content), mode="r:gz") as archive:
            member_names = archive.getnames()
            label_dirs = {
                "pos": "positive",
                "neg": "negative",
            }

            for folder_name, label in label_dirs.items():
                review_names = sorted(
                    name for name in member_names
                    if name.startswith(f"txt_sentoken/{folder_name}/") and name.endswith(".txt")
                )
                for review_name in review_names:
                    review_file = archive.extractfile(review_name)
                    if review_file is None:
                        continue
                    text = review_file.read().decode("latin-1").strip()
                    records.append({"label": label, "text": text})

        df = pd.DataFrame(records, columns=["label", "text"])
    except Exception as e:
        print(f"Error loading movie reviews data: {e}")
        print("Please ensure you have an internet connection.")
        raise

    return df

def load_mushroom_data():
    """
    Load the UCI Mushroom dataset.
    
    This dataset contains descriptions of hypothetical samples corresponding to 
    23 species of gilled mushrooms. Each sample is classified as either 
    poisonous (p) or edible (e).
    
    Returns:
    --------
    pd.DataFrame
        Mushroom dataset with 8124 samples and 23 features
        
    Examples:
    ---------
    >>> from cuanalytics.datasets import load_mushroom_data
    >>> df = load_mushroom_data()
    >>> print(df.shape)
    (8124, 23)
    >>> print(df['class'].value_counts())
    
    References:
    -----------
    https://archive.ics.uci.edu/dataset/73/mushroom
    """
    url = "https://archive.ics.uci.edu/ml/machine-learning-databases/mushroom/agaricus-lepiota.data"
    
    column_names = [
        'class', 'cap-shape', 'cap-surface', 'cap-color', 'bruises', 'odor',
        'gill-attachment', 'gill-spacing', 'gill-size', 'gill-color',
        'stalk-shape', 'stalk-root', 'stalk-surface-above-ring',
        'stalk-surface-below-ring', 'stalk-color-above-ring',
        'stalk-color-below-ring', 'veil-type', 'veil-color', 'ring-number',
        'ring-type', 'spore-print-color', 'population', 'habitat'
    ]
    
    try:
        # Try using requests (handles SSL better)
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        df = pd.read_csv(StringIO(response.text), names=column_names)
    except Exception as e:
        print(f"Error loading mushroom data: {e}")
        print("Please ensure you have an internet connection.")
        raise
    
    return df

def load_iris_data():
    """
    Load the classic Iris dataset.
    
    This dataset contains 150 samples of iris flowers with measurements
    of sepal and petal dimensions, classified into 3 species.
    
    Returns:
    --------
    pd.DataFrame
        Iris dataset with 150 samples and 5 columns
        
    Examples:
    ---------
    >>> from cuanalytics.datasets import load_iris_data
    >>> df = load_iris_data()
    >>> print(df['species'].value_counts())
    """
    url = "https://archive.ics.uci.edu/ml/machine-learning-databases/iris/iris.data"
    
    column_names = [
        'sepal_length', 'sepal_width', 'petal_length', 'petal_width', 'species'
    ]
    
    try:
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        df = pd.read_csv(StringIO(response.text), names=column_names)
        # Remove any empty rows
        df = df[df['species'].notna()]
    except Exception as e:
        print(f"Error loading iris data: {e}")
        raise
    
    return df

def load_breast_cancer_data():
    """
    Load the Breast Cancer Wisconsin (Diagnostic) dataset.
    
    This dataset contains features computed from digitized images of 
    fine needle aspirate (FNA) of breast masses. Features describe 
    characteristics of cell nuclei present in the images.
    
    The target variable is 'diagnosis':
    - 'M' = Malignant (cancerous)
    - 'B' = Benign (non-cancerous)
    
    Features include measurements like:
    - radius, texture, perimeter, area, smoothness
    - compactness, concavity, concave points, symmetry, fractal dimension
    
    Each feature has three versions: mean, standard error (se), and worst (largest).
    For example: radius_mean, radius_se, radius_worst
    
    Returns:
    --------
    df : pd.DataFrame
        DataFrame with 569 samples and 31 columns (30 features + diagnosis)
    
    Examples:
    ---------
    >>> from cuanalytics import load_breast_cancer_data, fit_lda, split_data
    >>> df = load_breast_cancer_data()
    >>> print(df.shape)
    (569, 31)
    >>> print(df['diagnosis'].value_counts())
    >>> train, test = split_data(df, test_size=0.2)
    >>> lda = fit_lda(train, formula='diagnosis ~ .')
    
    Dataset Information:
    --------------------
    Source: UCI Machine Learning Repository
    URL: https://archive.ics.uci.edu/dataset/17/breast+cancer+wisconsin+diagnostic
    Samples: 569
    Features: 30 (all numeric, real-valued)
    Classes: 2 (Malignant, Benign)
    Missing values: None
    
    Feature Groups (each has mean, se, worst):
    - radius: distances from center to points on perimeter
    - texture: standard deviation of gray-scale values
    - perimeter: size of core tumor
    - area: area of tumor
    - smoothness: local variation in radius lengths
    - compactness: perimeter^2 / area - 1.0
    - concavity: severity of concave portions of contour
    - concave points: number of concave portions of contour
    - symmetry: symmetry of the tumor
    - fractal dimension: "coastline approximation" - 1
    
    Citation:
    ---------
    William H. Wolberg, W. Nick Street, and Olvi L. Mangasarian.
    "Breast Cancer Wisconsin (Diagnostic) Data Set."
    UCI Machine Learning Repository, 1995.
    """
    from sklearn.datasets import load_breast_cancer
    import pandas as pd
    
    # Load from sklearn (which gets it from UCI)
    data = load_breast_cancer()
    
    # Create DataFrame with feature names
    df = pd.DataFrame(data.data, columns=data.feature_names)
    
    # Add target column
    # sklearn uses 0/1, convert to M/B for clarity
    df['diagnosis'] = ['M' if target == 0 else 'B' for target in data.target]
    
    return df



# Add to cuanalytics/datasets/loaders.py

def load_real_estate_data():
    """
    Load the Real Estate Valuation dataset from UCI.
    
    This dataset contains real estate valuation data from Sindian District,
    New Taipei City, Taiwan. The target is house price per unit area.
    
    Features:
    - transaction_date: Transaction date (e.g., 2013.250 = March 2013)
    - house_age: House age in years
    - distance_to_MRT: Distance to nearest MRT station in meters
    - num_convenience_stores: Number of convenience stores nearby
    - latitude: Latitude coordinate
    - longitude: Longitude coordinate
    
    Target:
    - price_per_unit: House price per unit area (10000 New Taiwan Dollar/Ping)
                      Note: 1 Ping = 3.3 square meters
    
    Returns:
    --------
    df : pd.DataFrame
        DataFrame with 414 samples and 7 columns (6 features + target)
    
    Examples:
    ---------
    >>> from cuanalytics import load_real_estate_data, fit_lm, split_data
    >>> df = load_real_estate_data()
    >>> print(df.shape)
    (414, 7)
    >>> print(df.head())
    >>> train, test = split_data(df, test_size=0.2)
    >>> model = fit_lm(train, formula='price_per_unit ~ .')
    >>> model.summary()
    
    Dataset Information:
    --------------------
    Source: UCI Machine Learning Repository
    URL: https://archive.ics.uci.edu/dataset/477/real+estate+valuation+data+set
    Location: Sindian District, New Taipei City, Taiwan
    Time period: 2012.667 - 2013.583 (August 2012 - July 2013)
    Samples: 414
    Features: 6 (all numeric)
    Target: Continuous (price per unit area)
    Missing values: None
    
    Citation:
    ---------
    Yeh, I-Cheng. (2018). Real Estate Valuation.
    UCI Machine Learning Repository.
    https://doi.org/10.24432/C5J30W
    """
    import pandas as pd
    import io
    import urllib.request
    import ssl
    
    # URL for the dataset
    url = "https://archive.ics.uci.edu/static/public/477/real+estate+valuation+data+set.zip"
    
    try:
        # Create SSL context that doesn't verify certificates
        ssl_context = ssl._create_unverified_context()
        
        print("Downloading Real Estate Valuation dataset from UCI...")
        
        # Download the zip file
        with urllib.request.urlopen(url, context=ssl_context) as response:
            zip_data = response.read()
        
        # Extract the Excel file from the zip
        import zipfile
        with zipfile.ZipFile(io.BytesIO(zip_data)) as zip_file:
            # The dataset is in an Excel file
            excel_filename = 'Real estate valuation data set.xlsx'
            with zip_file.open(excel_filename) as excel_file:
                # Read the Excel file (skip first row which is just a title)
                df = pd.read_excel(excel_file, header=1)
        
        print("Dataset loaded successfully!")
        
    except Exception as e:
        print(f"\nError downloading real estate data: {e}")
        print("\nFalling back to synthetic sales dataset...")
        print("If you need the real estate data, try downloading manually from:")
        print("https://archive.ics.uci.edu/dataset/477/real+estate+valuation+data+set")

        import numpy as np

        # Deterministic, numeric fallback matching expected schema.
        rng = np.random.default_rng(42)
        n_rows = 414

        transaction_date = rng.uniform(2012.667, 2013.583, size=n_rows)
        house_age = rng.uniform(0.0, 45.0, size=n_rows)
        distance_to_mrt = rng.uniform(20.0, 2000.0, size=n_rows)
        num_convenience = rng.integers(0, 11, size=n_rows).astype(float)
        latitude = rng.uniform(24.93, 25.15, size=n_rows)
        longitude = rng.uniform(121.40, 121.65, size=n_rows)

        # Synthetic target with reasonable signal and noise, always positive.
        base = 80.0
        noise = rng.normal(0.0, 5.0, size=n_rows)
        price_per_unit = (
            base
            - 0.5 * house_age
            - 0.01 * distance_to_mrt
            + 1.5 * num_convenience
            + noise
        )
        price_per_unit = np.clip(price_per_unit, 5.0, None)

        df = pd.DataFrame(
            {
                "transaction_date": transaction_date,
                "house_age": house_age,
                "distance_to_MRT": distance_to_mrt,
                "num_convenience_stores": num_convenience,
                "latitude": latitude,
                "longitude": longitude,
                "price_per_unit": price_per_unit,
            }
        )

        return df
    
    # Clean up column names
    df.columns = [
        'No',  # Row number (we'll drop this)
        'transaction_date',
        'house_age',
        'distance_to_MRT',
        'num_convenience_stores',
        'latitude',
        'longitude',
        'price_per_unit'
    ]
    
    # Drop the row number column
    df = df.drop('No', axis=1)
    
    # Reset index
    df = df.reset_index(drop=True)
    
    return df

# Module-level convenience dictionary
AVAILABLE_DATASETS = {
    'movie_reviews': load_movie_reviews_data,
    'sms_spam': load_sms_spam_data,
    'mushroom': load_mushroom_data,
    'breast_cancer': load_breast_cancer_data,
    'iris': load_iris_data,
    'real_estate': load_real_estate_data,
    # 'titanic': load_titanic_data,
    # 'sales': get_sample_sales_data,
}


def list_datasets():
    """
    List all available datasets in the cuanalytics package.
    
    Returns:
    --------
    list
        Names of available datasets
        
    Examples:
    ---------
    >>> from cuanalytics.datasets import list_datasets
    >>> print(list_datasets())
    """
    return list(AVAILABLE_DATASETS.keys())


def load_dataset(name):
    """
    Load a dataset by name.
    
    Parameters:
    -----------
    name : str
        Name of the dataset ('mushroom', etc.)
    
    Returns:
    --------
    pd.DataFrame
        Requested dataset
        
    Examples:
    ---------
    >>> from cuanalytics.datasets import load_dataset
    >>> df = load_dataset('mushroom')
    """
    if name not in AVAILABLE_DATASETS:
        available = ', '.join(AVAILABLE_DATASETS.keys())
        raise ValueError(f"Dataset '{name}' not found. Available datasets: {available}")
    
    return AVAILABLE_DATASETS[name]()



"""
Similarity and distance metrics.
"""

import numpy as np


def _to_array(x):
    return np.asarray(x, dtype=float)


def euclidean(a, b):
    """Compute Euclidean distance between two vectors."""
    a = _to_array(a)
    b = _to_array(b)
    return np.linalg.norm(a - b)


def manhattan(a, b):
    """Compute Manhattan (L1) distance between two vectors."""
    a = _to_array(a)
    b = _to_array(b)
    return np.sum(np.abs(a - b))


def cosine(a, b):
    """Compute cosine distance between two vectors."""
    a = _to_array(a)
    b = _to_array(b)
    denom = np.linalg.norm(a) * np.linalg.norm(b)
    if denom == 0:
        return 0.0
    return 1 - float(np.dot(a, b) / denom)


def jaccard(a, b):
    """Compute Jaccard distance between two binary vectors."""
    a = np.asarray(a).astype(bool)
    b = np.asarray(b).astype(bool)
    union = np.logical_or(a, b).sum()
    if union == 0:
        return 0.0
    intersection = np.logical_and(a, b).sum()
    return 1 - (intersection / union)



import numpy as np
import pandas as pd
import pytest

from cuanalytics import fit_nn
from cuanalytics.nn.neuralnet import NeuralNetModel


@pytest.fixture
def binary_data():
    np.random.seed(42)
    x1 = np.random.normal(0, 1, 120)
    x2 = np.random.normal(0, 1, 120)
    logits = 0.6 * x1 - 0.4 * x2
    probs = 1 / (1 + np.exp(-logits))
    y = np.where(probs > 0.5, 'A', 'B')
    return pd.DataFrame({'x1': x1, 'x2': x2, 'class': y})


@pytest.fixture
def regression_data():
    np.random.seed(7)
    x1 = np.random.normal(0, 1, 120)
    x2 = np.random.normal(0, 1, 120)
    y = 2.5 * x1 - 1.2 * x2 + np.random.normal(0, 0.1, 120)
    return pd.DataFrame({'x1': x1, 'x2': x2, 'y': y})


def test_fit_nn_returns_model(binary_data):
    model = fit_nn(binary_data, formula='class ~ x1 + x2', hidden_layers=[3, 2], solver='lbfgs', max_iter=1000)
    assert isinstance(model, NeuralNetModel)


def test_predict_returns_series(binary_data):
    model = fit_nn(binary_data, formula='class ~ x1 + x2', hidden_layers=[3], solver='lbfgs', max_iter=1000)
    preds = model.predict(binary_data)
    assert isinstance(preds, pd.Series)
    assert len(preds) == len(binary_data)


def test_score_classification(binary_data, monkeypatch):
    model = fit_nn(binary_data, formula='class ~ x1 + x2', hidden_layers=[4], solver='lbfgs', max_iter=1000)
    monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
    report = model.score(binary_data)
    assert isinstance(report, dict)
    assert 'accuracy' in report


def test_score_regression(regression_data):
    model = fit_nn(regression_data, formula='y ~ x1 + x2', hidden_layers=[4], solver='lbfgs', max_iter=1000)
    report = model.score(regression_data)
    assert 'r2' in report
    assert 'rmse' in report
    assert 'mae' in report


def test_predict_proba_classification(binary_data):
    model = fit_nn(binary_data, formula='class ~ x1 + x2', hidden_layers=[3], solver='lbfgs', max_iter=1000)
    probs = model.predict_proba(binary_data)
    assert probs.shape[0] == len(binary_data)
    assert probs.shape[1] == len(model.classes)


def test_predict_proba_regression_error(regression_data):
    model = fit_nn(regression_data, formula='y ~ x1 + x2', hidden_layers=[3], solver='lbfgs', max_iter=1000)
    with pytest.raises(ValueError, match="predict_proba is only available"):
        model.predict_proba(regression_data)


def test_get_metrics(binary_data):
    model = fit_nn(binary_data, formula='class ~ x1 + x2', hidden_layers=[3], solver='lbfgs', max_iter=1000)
    metrics = model.get_metrics()
    assert metrics['model_type'] == 'neural_network'
    assert metrics['task_type'] == 'classification'


def test_summary_returns_none(binary_data, monkeypatch):
    model = fit_nn(binary_data, formula='class ~ x1 + x2', hidden_layers=[3], solver='lbfgs', max_iter=1000)
    monkeypatch.setattr('builtins.print', lambda *args, **kwargs: None)
    result = model.summary()
    assert result is None


def test_visualize_runs(binary_data, monkeypatch):
    model = fit_nn(binary_data, formula='class ~ x1 + x2', hidden_layers=[3], solver='lbfgs', max_iter=1000)
    monkeypatch.setattr('matplotlib.pyplot.show', lambda: None)
    model.visualize()


def test_missing_formula_raises(binary_data):
    with pytest.raises(ValueError, match="Must provide 'formula'"):
        fit_nn(binary_data, formula=None)


def test_invalid_formula_raises(binary_data):
    with pytest.raises(ValueError, match="Formula must include a target"):
        fit_nn(binary_data, formula='class x1 + x2')


def test_missing_target_column_raises(binary_data):
    with pytest.raises(ValueError, match="Target 'missing' not found"):
        fit_nn(binary_data, formula='missing ~ x1 + x2')


def test_non_numeric_feature_raises(binary_data):
    df = binary_data.copy()
    df['cat'] = ['a', 'b'] * 60
    with pytest.raises(ValueError, match="All features must be numeric"):
        fit_nn(df, formula='class ~ x1 + cat')


def test_categorical_feature_with_C_allows_fit(binary_data):
    df = binary_data.copy()
    df['cat'] = ['a', 'b'] * 60
    model = fit_nn(df, formula='class ~ x1 + C(cat)', hidden_layers=[3], solver='lbfgs', max_iter=1000)
    assert isinstance(model, NeuralNetModel)


def test_unfitted_predict_raises(binary_data):
    model = NeuralNetModel.__new__(NeuralNetModel)
    with pytest.raises(RuntimeError, match="not been fitted"):
        model.predict(binary_data)


def test_transform_missing_model_spec_raises(binary_data):
    model = fit_nn(binary_data, formula='class ~ x1 + x2', hidden_layers=[3], solver='lbfgs', max_iter=1000)
    model.model_spec = None
    with pytest.raises(RuntimeError, match="Formula metadata missing"):
        model._transform_data_with_formula(binary_data)


def test_transform_rhs_attribute_branch(binary_data):
    model = fit_nn(binary_data, formula='class ~ x1 + x2', hidden_layers=[3], solver='lbfgs', max_iter=1000)

    class DummyMatrices:
        def __init__(self, rhs):
            self.rhs = rhs

    class DummySpec:
        def get_model_matrix(self, data, output='pandas'):
            rhs = pd.DataFrame(
                {
                    'Intercept': [1] * len(data),
                    'x1': data['x1'].values,
                    'x2': data['x2'].values
                },
                index=data.index
            )
            return DummyMatrices(rhs)

    model.model_spec = DummySpec()
    X_new = model._transform_data_with_formula(binary_data)
    assert list(X_new.columns) == ['x1', 'x2']



import pandas as pd
import pytest

from cuanalytics.formula import ModelFormula


def test_model_formula_supervised_builds_xy():
    df = pd.DataFrame({'x1': [1, 2], 'x2': [3, 4], 'y': [5, 6]})
    mf = ModelFormula(df, 'y ~ x1 + x2')
    assert mf.target == 'y'
    assert mf.y.tolist() == [5, 6]
    assert set(mf.feature_names) == {'x1', 'x2'}
    assert 'Intercept' not in mf.feature_names


def test_model_formula_unsupervised_rhs_only():
    df = pd.DataFrame({'x1': [1, 2], 'x2': [3, 4]})
    mf = ModelFormula(df, 'x1 + x2')
    assert mf.target is None
    assert mf.y is None
    assert set(mf.feature_names) == {'x1', 'x2'}


def test_model_formula_transform():
    df = pd.DataFrame({'x1': [1, 2], 'x2': [3, 4], 'y': [5, 6]})
    mf = ModelFormula(df, 'y ~ x1 + x2')
    new_df = pd.DataFrame({'x1': [10, 20], 'x2': [30, 40]})
    X_new = mf.transform(new_df)
    assert list(X_new.columns) == ['x1', 'x2']
    assert len(X_new) == 2


def test_model_formula_missing_target_raises():
    df = pd.DataFrame({'x1': [1, 2], 'x2': [3, 4]})
    with pytest.raises(ValueError, match="Target 'y' not found"):
        ModelFormula(df, 'y ~ x1 + x2')


def test_model_formula_validate_regression_requires_target():
    df = pd.DataFrame({'x1': [1, 2], 'x2': [3, 4]})
    mf = ModelFormula(df, 'x1 + x2')
    with pytest.raises(ValueError, match="Formula must include a target"):
        mf.validate_regression_target(df)



def test_model_formula_validate_regression_non_numeric_target():
    df = pd.DataFrame({'x1': [1, 2], 'y': ['a', 'b']})
    mf = ModelFormula(df, 'y ~ x1')
    with pytest.raises(ValueError, match="must be numeric"):
        mf.validate_regression_target(df)



"""
Text vectorization utilities for educational text mining workflows.
"""

import re

import numpy as np
import pandas as pd
from scipy.sparse import issparse
from sklearn.feature_extraction.text import CountVectorizer, ENGLISH_STOP_WORDS, TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import normalize


class TextVectorizer:
    """
    Fit and apply a bag-of-words style text representation.

    Parameters
    ----------
    text_col : str
        Column containing raw text.
    target_col : str | None
        Optional label column to append during transform().
    method : str
        'binary', 'count', 'tf', or 'tfidf'.
    max_features : int | None
        Maximum number of terms/ngrams to keep.
    min_df : int | float
        Minimum document frequency threshold.
    max_df : int | float
        Maximum document frequency threshold.
    max_n : int
        Include n-grams from size 1 through max_n.
    lowercase : bool
        Convert text to lowercase before tokenization.
    remove_stopwords : bool
        Remove common English stopwords.
    stem : bool
        Apply a lightweight educational stemmer.
    keep_numbers : bool
        If False, discard tokens that contain digits.
    """

    def __init__(
        self,
        text_col,
        target_col=None,
        method="tfidf",
        max_features=None,
        min_df=1,
        max_df=1.0,
        max_n=1,
        lowercase=True,
        remove_stopwords=True,
        stem=False,
        keep_numbers=False,
    ):
        self.text_col = text_col
        self.target_col = target_col
        self.method = method
        self.max_features = max_features
        self.min_df = min_df
        self.max_df = max_df
        self.max_n = max_n
        self.lowercase = lowercase
        self.remove_stopwords = remove_stopwords
        self.stem = stem
        self.keep_numbers = keep_numbers

        self.vectorizer = None
        self.feature_names_ = None
        self.output_columns_ = None
        self.vocabulary_ = None
        self._corpus_matrix = None
        self._fitted_index = None
        self._fitted_texts = None

    def _validate_params(self):
        allowed_methods = {"binary", "count", "tf", "tfidf"}
        if self.method not in allowed_methods:
            raise ValueError(f"method must be one of {sorted(allowed_methods)}")
        if not isinstance(self.max_n, int) or self.max_n < 1:
            raise ValueError("max_n must be an integer >= 1")

    def _validate_input_df(self, df):
        if not isinstance(df, pd.DataFrame):
            raise TypeError("df must be a pandas DataFrame")
        if self.text_col not in df.columns:
            raise ValueError(f"Text column '{self.text_col}' not found in data")

    def _normalize_text(self, text):
        if pd.isna(text):
            return ""
        text = str(text)
        if self.lowercase:
            text = text.lower()
        return text

    def _simple_stem(self, token):
        """
        Lightweight suffix stripping for educational use.
        """
        replacements = [
            ("ingly", ""),
            ("edly", ""),
            ("ing", ""),
            ("edly", ""),
            ("ed", ""),
            ("ies", "y"),
            ("sses", "ss"),
            ("s", ""),
        ]
        for suffix, replacement in replacements:
            if token.endswith(suffix) and len(token) > len(suffix) + 2:
                candidate = token[: -len(suffix)] + replacement
                if len(candidate) >= 3:
                    return candidate
        return token

    def _tokenize(self, doc):
        doc = self._normalize_text(doc)
        tokens = re.findall(r"\b\w+\b", doc)
        if not self.keep_numbers:
            tokens = [token for token in tokens if not any(ch.isdigit() for ch in token)]
        tokens = [
            token for token in tokens
            if len(token) > 1 or any(ch.isdigit() for ch in token)
        ]
        if self.remove_stopwords:
            tokens = [token for token in tokens if token not in ENGLISH_STOP_WORDS]
        if self.stem:
            tokens = [self._simple_stem(token) for token in tokens]
        return [token for token in tokens if token]

    def _create_vectorizer(self):
        common_kwargs = {
            "preprocessor": self._normalize_text,
            "tokenizer": self._tokenize,
            "token_pattern": None,
            "ngram_range": (1, self.max_n),
            "max_features": self.max_features,
            "min_df": self.min_df,
            "max_df": self.max_df,
        }
        if self.method == "tfidf":
            return TfidfVectorizer(norm="l2", **common_kwargs)
        if self.method == "binary":
            return CountVectorizer(binary=True, **common_kwargs)
        return CountVectorizer(binary=False, **common_kwargs)

    def _build_output_columns(self, feature_names):
        used = set()
        output_columns = []
        for feature in feature_names:
            safe = re.sub(r"[^0-9A-Za-z_]+", "_", feature).strip("_")
            if not safe:
                safe = "term"
            safe = f"term__{safe}"
            candidate = safe
            counter = 2
            while candidate in used:
                candidate = f"{safe}_{counter}"
                counter += 1
            used.add(candidate)
            output_columns.append(candidate)
        return output_columns

    def fit(self, df):
        """
        Learn vocabulary and weighting statistics from a training DataFrame.
        """
        self._validate_params()
        self._validate_input_df(df)

        self.vectorizer = self._create_vectorizer()
        texts = df[self.text_col].fillna("").astype(str)
        matrix = self.vectorizer.fit_transform(texts)
        if self.method == "tf":
            matrix = normalize(matrix, norm="l1", axis=1)

        self.feature_names_ = list(self.vectorizer.get_feature_names_out())
        self.output_columns_ = self._build_output_columns(self.feature_names_)
        self.vocabulary_ = dict(zip(self.feature_names_, self.output_columns_))
        self._corpus_matrix = matrix
        self._fitted_index = df.index.copy()
        self._fitted_texts = texts.reset_index(drop=True).copy()
        return self

    def _check_fitted(self):
        if self.vectorizer is None or self.feature_names_ is None:
            raise RuntimeError(
                "TextVectorizer has not been fitted yet. "
                "Create one with: vec = fit_text_vectorizer(df, text_col='review')"
            )

    def _matrix_to_dataframe(self, matrix, index):
        if issparse(matrix):
            matrix = matrix.toarray()
        return pd.DataFrame(matrix, index=index, columns=self.output_columns_)

    def transform_text(self, texts):
        """
        Transform raw text values into a feature DataFrame.
        """
        self._check_fitted()

        if isinstance(texts, str):
            texts = [texts]

        index = pd.RangeIndex(start=0, stop=len(texts))
        matrix = self.vectorizer.transform(pd.Series(list(texts), dtype="object"))
        if self.method == "tf":
            matrix = normalize(matrix, norm="l1", axis=1)
        return self._matrix_to_dataframe(matrix, index=index)

    def transform_query(self, text):
        """
        Transform a single query string into a one-row feature DataFrame.
        """
        return self.transform_text([text])

    def transform(self, df, include_target=True):
        """
        Transform a DataFrame containing raw text into text features.
        """
        self._validate_input_df(df)
        transformed = self.transform_text(df[self.text_col].fillna("").astype(str).tolist())
        transformed.index = df.index

        if include_target and self.target_col and self.target_col in df.columns:
            transformed[self.target_col] = df[self.target_col]
        return transformed

    def fit_transform(self, df, include_target=True):
        """
        Fit on df, then transform it.
        """
        self.fit(df)
        return self.transform(df, include_target=include_target)

    def get_feature_names(self):
        """
        Return the learned terms/ngrams in vocabulary order.
        """
        self._check_fitted()
        return list(self.feature_names_)

    def summary(self):
        """
        Print a short summary of the fitted vectorizer.
        """
        self._check_fitted()
        print("\nText Vectorizer fitted successfully!")
        print(f"  Text column: {self.text_col}")
        print(f"  Method: {self.method}")
        print(f"  Max n-gram size: {self.max_n}")
        print(f"  Vocabulary size: {len(self.feature_names_)}")
        print(f"  Stopwords removed: {self.remove_stopwords}")
        print(f"  Stemming enabled: {self.stem}")

    def top_terms(self, df=None, top_n=20, by="frequency"):
        """
        Return the most heavily weighted terms in the fitted corpus or a provided DataFrame.
        """
        self._check_fitted()
        if by not in {"frequency", "tfidf"}:
            raise ValueError("by must be 'frequency' or 'tfidf'")

        if df is None:
            texts = self._fitted_texts
            matrix = self._corpus_matrix
        else:
            texts = df[self.text_col].fillna("").astype(str)
            if by == "tfidf":
                matrix = self.vectorizer.transform(texts)
                if self.method == "tf":
                    matrix = normalize(matrix, norm="l1", axis=1)

        if by == "frequency":
            count_vectorizer = CountVectorizer(
                vocabulary=self.vectorizer.vocabulary_,
                preprocessor=self._normalize_text,
                tokenizer=self._tokenize,
                token_pattern=None,
                ngram_range=(1, self.max_n),
            )
            count_matrix = count_vectorizer.transform(texts)
            scores = np.asarray(count_matrix.sum(axis=0)).ravel()
        else:
            scores = np.asarray(matrix.sum(axis=0)).ravel()

        results = pd.DataFrame({"term": self.feature_names_, "score": scores})
        return results.sort_values("score", ascending=False).head(top_n).reset_index(drop=True)

    def top_terms_by_class(self, df, top_n=20):
        """
        Rank terms by information gain with respect to the configured target column.
        """
        self._check_fitted()
        if self.target_col is None:
            raise ValueError("target_col must be set to compute class-based term rankings.")
        if self.target_col not in df.columns:
            raise ValueError(f"Target column '{self.target_col}' not found in data")

        transformed = self.transform(df, include_target=False)
        y = df[self.target_col]

        from sklearn.feature_selection import mutual_info_classif

        scores = mutual_info_classif(transformed.values, y)
        results = pd.DataFrame({"term": self.feature_names_, "information_gain": scores})
        return results.sort_values("information_gain", ascending=False).head(top_n).reset_index(drop=True)

    def similarity_search(self, df, query, top_n=5):
        """
        Return the most similar documents in df to a query string.
        """
        self._validate_input_df(df)
        document_features = self.transform(df, include_target=False)
        query_features = self.transform_query(query)
        scores = cosine_similarity(query_features.values, document_features.values).ravel()

        results = df.copy()
        results["similarity"] = scores
        return results.sort_values("similarity", ascending=False).head(top_n)


def fit_text_vectorizer(
    df,
    text_col,
    target_col=None,
    method="tfidf",
    max_features=None,
    min_df=1,
    max_df=1.0,
    max_n=1,
    lowercase=True,
    remove_stopwords=True,
    stem=False,
    keep_numbers=False,
):
    """
    Fit and return a TextVectorizer.
    """
    vectorizer = TextVectorizer(
        text_col=text_col,
        target_col=target_col,
        method=method,
        max_features=max_features,
        min_df=min_df,
        max_df=max_df,
        max_n=max_n,
        lowercase=lowercase,
        remove_stopwords=remove_stopwords,
        stem=stem,
        keep_numbers=keep_numbers,
    )
    return vectorizer.fit(df)



"""Confusion matrix utilities for classification models."""

import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix


class ConfusionMatrix:
    """Represent a confusion matrix and derived classification metrics."""

    def __init__(self, y_true, y_pred, labels=None, positive_label=None):
        if labels is None:
            labels = list(pd.unique(np.concatenate([np.asarray(y_true), np.asarray(y_pred)])))
        self.labels = list(labels)
        if not self.labels:
            raise ValueError("labels cannot be empty")

        self.matrix = confusion_matrix(y_true, y_pred, labels=self.labels)
        self.total = int(self.matrix.sum())

        if positive_label is None and len(self.labels) == 2:
            positive_label = self.labels[1]
        if positive_label is not None and positive_label not in self.labels:
            raise ValueError(f"positive_label '{positive_label}' not found in labels")
        self.positive_label = positive_label

    @classmethod
    def from_matrix(cls, matrix, labels, positive_label=None):
        """Create a ConfusionMatrix instance from an existing matrix."""
        matrix = np.asarray(matrix)
        labels = list(labels)
        if matrix.ndim != 2 or matrix.shape[0] != matrix.shape[1]:
            raise ValueError("matrix must be a square 2D array")
        if matrix.shape[0] != len(labels):
            raise ValueError(
                f"labels length must match matrix size ({matrix.shape[0]}), got {len(labels)}"
            )

        obj = cls.__new__(cls)
        obj.labels = labels
        obj.matrix = matrix
        obj.total = int(matrix.sum())
        if positive_label is None and len(labels) == 2:
            positive_label = labels[1]
        if positive_label is not None and positive_label not in labels:
            raise ValueError(f"positive_label '{positive_label}' not found in labels")
        obj.positive_label = positive_label
        return obj

    def to_dataframe(self, display='inverted'):
        """Return a labeled confusion matrix table.

        display='inverted' (default): textbook binary orientation with
        rows=predicted, cols=actual.
        display='normal': rows=actual, cols=predicted.
        """
        if display not in {'normal', 'inverted'}:
            raise ValueError("display must be 'normal' or 'inverted'")

        ordered_matrix, ordered_labels = self._matrix_for_display(display)

        if display == 'normal':
            return pd.DataFrame(
                ordered_matrix,
                index=[f"Actual {label}" for label in ordered_labels],
                columns=[f"Pred {label}" for label in ordered_labels],
            )

        return pd.DataFrame(
            ordered_matrix.T,
            index=[f"Pred {label}" for label in ordered_labels],
            columns=[f"Actual {label}" for label in ordered_labels],
        )

    def summary(self, display='inverted'):
        """Print a readable confusion matrix report and return the table."""
        metrics = self.get_metrics()
        conf_df = self.to_dataframe(display=display)
        print("\nConfusion Matrix and Statistics")
        print("=" * 60)
        print("\nConfusion Matrix:")
        print(conf_df.to_string())

        overall = {
            'Accuracy': metrics['accuracy'],
            'Kappa': metrics['kappa'],
            'Macro Precision': metrics['macro_precision'],
            'Macro Recall': metrics['macro_recall'],
            'Macro Specificity': metrics['macro_specificity'],
            'Macro F1': metrics['macro_f1'],
        }

        if len(self.labels) == 2:
            overall.update({
                'Precision': metrics['precision'],
                'Recall': metrics['recall'],
                'Sensitivity': metrics['sensitivity'],
                'Specificity': metrics['specificity'],
                'FPR': metrics['fpr'],
                'FNR': metrics['fnr'],
                'F-measure': metrics['f_measure'],
            })

        print("\nOverall Statistics:")
        print(pd.Series(overall).to_string(float_format=lambda x: f"{x:.4f}"))

        if len(self.labels) == 2:
            counts = metrics['binary_counts']
            print(f"\nBinary Counts (Positive = '{metrics['positive_label']}'):")
            print(f"TP: {counts['tp']}  FP: {counts['fp']}  FN: {counts['fn']}  TN: {counts['tn']}")

        per_class_df = pd.DataFrame(metrics['per_class']).T
        print("\nStatistics by Class:")
        print(per_class_df.to_string(float_format=lambda x: f"{x:.4f}"))

        return conf_df

    def get_metrics(self):
        """Return confusion-matrix derived metrics as a dictionary."""
        cm = self.matrix
        total = self.total
        accuracy = np.trace(cm) / total if total else 0.0

        row_marginals = cm.sum(axis=1)
        col_marginals = cm.sum(axis=0)
        expected = (row_marginals * col_marginals).sum() / (total ** 2) if total else 0.0
        kappa = (accuracy - expected) / (1 - expected) if (1 - expected) else 0.0

        per_class = {}
        for idx, label in enumerate(self.labels):
            tp = cm[idx, idx]
            fn = cm[idx, :].sum() - tp
            fp = cm[:, idx].sum() - tp
            tn = total - tp - fn - fp

            precision = tp / (tp + fp) if (tp + fp) else 0.0
            recall = tp / (tp + fn) if (tp + fn) else 0.0
            specificity = tn / (tn + fp) if (tn + fp) else 0.0
            f1 = (2 * precision * recall / (precision + recall)) if (precision + recall) else 0.0

            per_class[label] = {
                'precision': precision,
                'recall': recall,
                'sensitivity': recall,
                'specificity': specificity,
                'f1': f1,
            }

        macro_precision = float(np.mean([m['precision'] for m in per_class.values()]))
        macro_recall = float(np.mean([m['recall'] for m in per_class.values()]))
        macro_specificity = float(np.mean([m['specificity'] for m in per_class.values()]))
        macro_f1 = float(np.mean([m['f1'] for m in per_class.values()]))

        metrics = {
            'accuracy': accuracy,
            'kappa': kappa,
            'confusion_matrix': cm,
            'classes': self.labels,
            'per_class': per_class,
            'macro_precision': macro_precision,
            'macro_recall': macro_recall,
            'macro_specificity': macro_specificity,
            'macro_f1': macro_f1,
        }

        if len(self.labels) == 2:
            positive_label = self.positive_label if self.positive_label is not None else self.labels[1]
            pos_idx = self.labels.index(positive_label)

            tp = cm[pos_idx, pos_idx]
            fn = cm[pos_idx, :].sum() - tp
            fp = cm[:, pos_idx].sum() - tp
            tn = total - tp - fn - fp

            sensitivity = tp / (tp + fn) if (tp + fn) else 0.0
            specificity = tn / (tn + fp) if (tn + fp) else 0.0
            precision = tp / (tp + fp) if (tp + fp) else 0.0
            recall = sensitivity
            tpr = sensitivity
            fpr = fp / (fp + tn) if (fp + tn) else 0.0
            fnr = fn / (fn + tp) if (fn + tp) else 0.0
            tnr = specificity
            f_measure = (2 * precision * recall / (precision + recall)) if (precision + recall) else 0.0

            metrics.update({
                'positive_label': positive_label,
                'binary_counts': {'tp': tp, 'fp': fp, 'fn': fn, 'tn': tn},
                'sensitivity': sensitivity,
                'specificity': specificity,
                'recall': recall,
                'precision': precision,
                'tpr': tpr,
                'fpr': fpr,
                'fnr': fnr,
                'tnr': tnr,
                'f_measure': f_measure,
                'f1': f_measure,
            })

        return metrics

    def get_expected_value(self, costs, display='inverted'):
        """Return expected value per observation for a given cost/benefit matrix.

        Costs are interpreted with the same layout as `to_dataframe(display=...)`.
        Default is `display='inverted'`.
        """
        costs = np.asarray(costs, dtype=float)
        matrix_for_costs = self.to_dataframe(display=display).to_numpy()
        if costs.shape != matrix_for_costs.shape:
            raise ValueError(
                f"costs must have shape {matrix_for_costs.shape}, got {costs.shape}"
            )

        total_value = float(np.sum(matrix_for_costs * costs))
        return total_value / self.total if self.total else 0.0

    def _matrix_for_display(self, display):
        """Return matrix and labels in the requested display order."""
        if display not in {'normal', 'inverted'}:
            raise ValueError("display must be 'normal' or 'inverted'")

        ordered_labels = self.labels
        ordered_matrix = self.matrix
        if display == 'inverted' and len(self.labels) == 2:
            # Textbook binary layout with positive class first:
            # [[TP, FP],
            #  [FN, TN]]
            positive_label = self.positive_label if self.positive_label is not None else self.labels[1]
            negative_label = next(label for label in self.labels if label != positive_label)
            ordered_labels = [positive_label, negative_label]
            idx = [self.labels.index(label) for label in ordered_labels]
            ordered_matrix = self.matrix[np.ix_(idx, idx)]
        return ordered_matrix, ordered_labels



# cuanalytics/clustering/kmeans.py
"""
K-Means clustering.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import silhouette_score
from sklearn.decomposition import PCA


class KMeansModel:
    """
    K-Means clustering model.
    """

    def __init__(self, df, formula, n_clusters=3, random_state=None, n_init='auto'):
        self.original_df = df
        self.formula = formula
        self.n_clusters = n_clusters
        self.random_state = random_state
        self.n_init = n_init
        self.model_spec = None

        if formula is None:
            raise ValueError("Must provide 'formula' for model specification")

        try:
            from formulaic import model_matrix
        except ImportError:
            raise ImportError(
                "Formula support requires the 'formulaic' library.\n"
                "Install it with: pip install formulaic"
            )

        rhs = formula
        if '~' in formula:
            rhs = formula.split('~', 1)[1].strip()

        df_rhs = df.copy()
        model_matrices = model_matrix(rhs, df_rhs, output='pandas')

        if hasattr(model_matrices, 'rhs'):
            X = model_matrices.rhs
        else:
            X = model_matrices

        self.model_spec = getattr(model_matrices, 'model_spec', None)

        if 'Intercept' in X.columns:
            X = X.drop('Intercept', axis=1)

        self.X = X
        self.feature_names = list(self.X.columns)

        used_vars = set()
        if self.model_spec is not None:
            rhs_spec = getattr(self.model_spec, 'rhs', self.model_spec)
            used_vars = {var for var in getattr(rhs_spec, 'variables', set())
                         if var in df_rhs.columns}
        if not used_vars:
            used_vars = set(df_rhs.columns)
        non_numeric = [col for col in used_vars
                       if not pd.api.types.is_numeric_dtype(df_rhs[col])]
        if non_numeric:
            raise ValueError("All features must be numeric. Encode categorical variables first.")

        self._fit()
        self._print_fit_summary()

    def _fit(self):
        self.model = KMeans(
            n_clusters=self.n_clusters,
            random_state=self.random_state,
            n_init=self.n_init
        )
        self.model.fit(self.X)
        self.labels_ = self.model.labels_

    def _check_fitted(self):
        if not hasattr(self, 'model') or self.model is None:
            raise RuntimeError(
                "Model has not been fitted yet. "
                "Create model with: model = fit_kmeans(df, formula='x1 + x2')"
            )

    def _print_fit_summary(self):
        print("\nK-Means fitted successfully!")
        print(f"  Clusters: {self.n_clusters}")
        print(f"  Features: {len(self.feature_names)}")
        print(f"  Samples: {len(self.X)}")

    def _transform_data_with_formula(self, df):
        if getattr(self, 'model_spec', None) is None:
            raise RuntimeError("Formula metadata missing; cannot transform new data.")

        rhs_spec = getattr(self.model_spec, 'rhs', self.model_spec)
        model_matrices = rhs_spec.get_model_matrix(df, output='pandas')

        if hasattr(model_matrices, 'rhs'):
            X_new = model_matrices.rhs
        else:
            X_new = model_matrices

        if 'Intercept' in X_new.columns:
            X_new = X_new.drop('Intercept', axis=1)

        return X_new

    def predict(self, df):
        self._check_fitted()
        X = self._transform_data_with_formula(df)
        labels = self.model.predict(X)
        return pd.Series(labels, index=df.index, name='cluster')

    def score(self, df):
        self._check_fitted()
        X = self._transform_data_with_formula(df)
        labels = self.model.predict(X)
        metrics = {'inertia': float(self.model.inertia_)}
        if self.n_clusters > 1 and len(X) > self.n_clusters:
            metrics['silhouette'] = float(silhouette_score(X, labels))
        return metrics

    def get_metrics(self):
        self._check_fitted()
        metrics = {
            'model_type': 'kmeans',
            'n_clusters': self.n_clusters,
            'n_samples': len(self.X),
            'n_features': len(self.feature_names),
            'feature_names': self.feature_names,
            'inertia': float(self.model.inertia_),
            'cluster_counts': {
                int(cluster): int(count)
                for cluster, count in pd.Series(self.labels_).value_counts().sort_index().items()
            },
        }
        if self.n_clusters > 1 and len(self.X) > self.n_clusters:
            metrics['silhouette'] = float(silhouette_score(self.X, self.labels_))
        return metrics

    def _extract_condensed_positive_rule(self, tree, feature_names):
        tree_ = tree.tree_

        def recurse(node, path, candidates):
            feature_idx = tree_.feature[node]
            if feature_idx != -2:
                threshold = tree_.threshold[node]
                feature = feature_names[feature_idx]
                left_path = path + [f"{feature} <= {threshold:.3f}"]
                right_path = path + [f"{feature} > {threshold:.3f}"]
                recurse(tree_.children_left[node], left_path, candidates)
                recurse(tree_.children_right[node], right_path, candidates)
                return

            counts = tree_.value[node][0]
            predicted = int(np.argmax(counts))
            positive_support = int(counts[1]) if len(counts) > 1 else 0
            if predicted == 1 and positive_support > 0:
                candidates.append((positive_support, path))

        candidates = []
        recurse(0, [], candidates)
        if not candidates:
            return "No concise rule found at this tree depth."

        best_path = max(candidates, key=lambda x: x[0])[1]
        if not best_path:
            return "All observations (root rule)."
        return " AND ".join(best_path)

    def describe_clusters(self, max_depth=3, criterion='entropy', random_state=42):
        """
        Return original training data with cluster labels and condensed rule descriptions.

        For each cluster, this fits a one-vs-rest decision tree on the model features and
        extracts a concise positive rule. Each row receives the rule for its assigned cluster.
        """
        self._check_fitted()

        allowed_criteria = {'entropy', 'gini'}
        if criterion not in allowed_criteria:
            raise ValueError("criterion must be 'entropy' or 'gini'.")

        unique_clusters = np.sort(np.unique(self.labels_))
        cluster_rules = {}
        X_values = self.X.values

        for cluster in unique_clusters:
            y_binary = (self.labels_ == cluster).astype(int)
            tree = DecisionTreeClassifier(
                max_depth=max_depth,
                criterion=criterion,
                random_state=random_state,
            )
            tree.fit(X_values, y_binary)
            cluster_rules[int(cluster)] = self._extract_condensed_positive_rule(tree, self.feature_names)

        labeled = self.original_df.copy()
        labels_series = pd.Series(self.labels_, index=self.X.index, name='cluster').astype(int)
        labeled['cluster'] = labels_series.reindex(labeled.index)
        labeled['cluster_rule'] = labeled['cluster'].map(cluster_rules)
        return labeled

    def visualize(self, figsize=(8, 6)):
        self._check_fitted()
        if len(self.feature_names) > 2:
            pca = PCA(n_components=2)
            X_plot = pca.fit_transform(self.X)
            xlabel = 'PC1'
            ylabel = 'PC2'
        else:
            X_plot = self.X.values
            xlabel, ylabel = self.feature_names

        unique_labels = np.sort(np.unique(self.labels_))
        base_colors = [
            '#ff7f00', '#377eb8', '#4daf4a', '#984ea3', '#b8860b',
            '#a65628', '#f781bf', '#999999', '#66c2a5', '#1b9e77',
        ]
        marker_cycle = ['o', 's', '^', 'D', 'P', 'X', 'v', '<', '>', '*']
        colors = {label: base_colors[i % len(base_colors)] for i, label in enumerate(unique_labels)}
        markers = {label: marker_cycle[i % len(marker_cycle)] for i, label in enumerate(unique_labels)}

        fig, ax = plt.subplots(figsize=figsize)
        for label in unique_labels:
            mask = self.labels_ == label
            ax.scatter(
                X_plot[mask, 0],
                X_plot[mask, 1],
                color=colors[label],
                marker=markers[label],
                alpha=0.9,
                edgecolors='white',
                linewidth=0.7,
            )
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title('K-Means Clusters')
        ax.grid(True, alpha=0.3)
        plt.tight_layout()
        plt.show()

    def summary(self):
        self._check_fitted()
        metrics = self.get_metrics()
        print("\n" + "=" * 70)
        print("K-MEANS SUMMARY")
        print("=" * 70)
        print(f"Clusters: {self.n_clusters}")
        print(f"Features: {len(self.feature_names)}")
        print(f"Samples: {len(self.X)}")
        print(f"Inertia: {metrics['inertia']:.4f}")
        if 'silhouette' in metrics:
            print(f"Silhouette: {metrics['silhouette']:.4f}")
        print("\nCluster Counts:")
        print(metrics['cluster_counts'])
        print("\n" + "=" * 70)


def fit_kmeans(df, formula, n_clusters=3, random_state=None, n_init='auto'):
    if formula is None:
        raise ValueError("Must provide 'formula' for model specification")
    return KMeansModel(
        df,
        formula=formula,
        n_clusters=n_clusters,
        random_state=random_state,
        n_init=n_init,
    )



# cuanalytics/knn/classifier.py
"""
k-Nearest Neighbors classifier.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.neighbors import KNeighborsClassifier
from cuanalytics.formula import allow_categorical_from_formula
from cuanalytics.metrics import ConfusionMatrix


class KNNClassifierModel:
    """
    K-Nearest Neighbors model for classification.
    """

    def __init__(self, df, formula, k=5, weights='uniform', metric='minkowski'):
        self.original_df = df
        self.formula = formula
        self.k = k
        self.weights = weights
        self.metric = metric
        self.model_spec = None

        if formula is None:
            raise ValueError("Must provide 'formula' for model specification")

        try:
            from formulaic import model_matrix
            from formulaic.transforms import C as categorical
        except ImportError:
            raise ImportError(
                "Formula support requires the 'formulaic' library.\n"
                "Install it with: pip install formulaic"
            )

        if '~' not in formula:
            raise ValueError("Formula must include a target (e.g., 'class ~ x1 + x2').")

        lhs, rhs = formula.split('~', 1)
        lhs = lhs.strip()
        rhs = rhs.strip()

        if lhs and lhs not in df.columns and lhs.isidentifier():
            raise ValueError(f"Target '{lhs}' not found in data")

        y = df[lhs]
        df_rhs = df.drop(columns=[lhs])
        try:
            model_matrices = model_matrix(rhs, df_rhs, output='pandas', context={"C": categorical})
        except TypeError:
            model_matrices = model_matrix(rhs, df_rhs, output='pandas')

        if hasattr(model_matrices, 'rhs'):
            X = model_matrices.rhs
        else:
            X = model_matrices

        self.model_spec = getattr(model_matrices, 'model_spec', None)
        self.target = lhs

        if 'Intercept' in X.columns:
            X = X.drop('Intercept', axis=1)

        self.df = X.copy()
        self.df[self.target] = y
        self.X = X
        self.y = y

        self.feature_names = list(self.X.columns)
        self.feature_names_original = [col for col in df.columns if col != self.target]
        self.classes = sorted(self.y.unique())

        used_vars = set()
        if self.model_spec is not None:
            rhs_spec = getattr(self.model_spec, 'rhs', self.model_spec)
            used_vars = {var for var in getattr(rhs_spec, 'variables', set())
                         if var in df_rhs.columns}
        if not used_vars:
            used_vars = set(df_rhs.columns)
        non_numeric = [col for col in used_vars
                       if not pd.api.types.is_numeric_dtype(df_rhs[col])]
        non_numeric = allow_categorical_from_formula(non_numeric, self.formula)
        if non_numeric:
            raise ValueError(
                "All features must be numeric unless wrapped in C(). "
                "Encode categorical variables or use C() in the formula."
            )

        self._fit()
        self._print_fit_summary()

    def _fit(self):
        self.model = KNeighborsClassifier(n_neighbors=self.k, weights=self.weights, metric=self.metric)
        self.model.fit(self.X, self.y)

    def _check_fitted(self):
        if not hasattr(self, 'model') or self.model is None:
            raise RuntimeError(
                "Model has not been fitted yet. "
                "Create model with: model = fit_knn_classifier(df, formula='class ~ x1 + x2')"
            )

    def _print_fit_summary(self):
        print("\nKNN Classifier fitted successfully!")
        print(f"  Classes: {self.classes}")
        print(f"  Features: {len(self.feature_names)}")
        print(f"  Training samples: {len(self.df)}")
        print(f"  k: {self.k}")

    def _transform_data_with_formula(self, df):
        if getattr(self, 'model_spec', None) is None:
            raise RuntimeError("Formula metadata missing; cannot transform new data.")

        rhs_spec = getattr(self.model_spec, 'rhs', self.model_spec)
        df_rhs = df.drop(columns=[self.target], errors='ignore')
        model_matrices = rhs_spec.get_model_matrix(df_rhs, output='pandas')

        if hasattr(model_matrices, 'rhs'):
            X_new = model_matrices.rhs
        else:
            X_new = model_matrices

        if 'Intercept' in X_new.columns:
            X_new = X_new.drop('Intercept', axis=1)

        return X_new

    def _compute_classification_metrics(self, y_true, y_pred):
        cm = ConfusionMatrix(y_true, y_pred, labels=self.classes)
        return cm.get_metrics()

    def _print_score_report(self, report, display='inverted'):
        print("\nSCORE REPORT")
        print("=" * 60)
        print(f"Accuracy: {report['accuracy']:.2%}")
        print(f"Kappa: {report['kappa']:.4f}")

        conf_obj = ConfusionMatrix.from_matrix(report['confusion_matrix'], labels=self.classes)
        conf_df = conf_obj.to_dataframe(display=display)
        print("\nConfusion Matrix:")
        print(conf_df.to_string())

        per_class_df = pd.DataFrame(report['per_class']).T
        print("\nPer-Class Metrics:")
        print(per_class_df.to_string(float_format=lambda x: f"{x:.4f}"))

    def predict(self, df):
        self._check_fitted()
        X = self._transform_data_with_formula(df)
        preds = self.model.predict(X)
        return pd.Series(preds, index=df.index, name=self.target)

    def predict_proba(self, df):
        """
        Predict class probabilities for new data.
        """
        self._check_fitted()
        X = self._transform_data_with_formula(df)
        return self.model.predict_proba(X)

    def _resolve_positive_class(self, positive_class=None):
        classes = list(self.classes)
        if len(classes) != 2:
            return None
        if positive_class is not None:
            if positive_class not in classes:
                raise ValueError(f"positive_class '{positive_class}' not found in classes {classes}.")
            return positive_class
        class_set = set(classes)
        if class_set == {0, 1}:
            return 1
        if class_set == {"0", "1"}:
            return "1"
        return sorted(classes)[-1]

    def score(self, df, display='inverted', threshold=0.5, positive_class=None):
        report = self.get_score(df, threshold=threshold, positive_class=positive_class)
        self._print_score_report(report, display=display)
        return report

    def get_score(self, df, threshold=0.5, positive_class=None):
        self._check_fitted()
        X = self._transform_data_with_formula(df)
        y_true = df[self.target]

        if not 0.0 <= float(threshold) <= 1.0:
            raise ValueError("threshold must be between 0 and 1.")

        model_classes = list(self.model.classes_)
        if len(model_classes) == 2:
            pos_class = self._resolve_positive_class(positive_class=positive_class)
            pos_idx = model_classes.index(pos_class)
            neg_class = next(c for c in model_classes if c != pos_class)
            proba = self.model.predict_proba(X)[:, pos_idx]
            y_pred = np.where(proba >= float(threshold), pos_class, neg_class)
        else:
            if positive_class is not None:
                raise ValueError(
                    "positive_class and threshold-based scoring are only supported for binary classification."
                )
            y_pred = self.model.predict(X)

        report = self._compute_classification_metrics(y_true, y_pred)
        return report

    def get_confusion_matrix(self, df=None):
        self._check_fitted()
        if df is None:
            y_true = self.y
            y_pred = self.model.predict(self.X)
        else:
            X = self._transform_data_with_formula(df)
            y_true = df[self.target]
            y_pred = self.model.predict(X)
        return ConfusionMatrix(y_true, y_pred, labels=self.classes)

    def get_metrics(self):
        self._check_fitted()
        y_pred = self.model.predict(self.X)
        report = self._compute_classification_metrics(self.y, y_pred)
        return {
            'model_type': 'knn_classifier',
            'target': self.target,
            'n_samples': len(self.y),
            'n_features': len(self.feature_names),
            'feature_names': self.feature_names,
            'classes': self.classes,
            'k': self.k,
            'weights': self.weights,
            'metric': self.metric,
            'score': report,
        }

    def visualize(self, figsize=(8, 6)):
        self._check_fitted()
        if len(self.feature_names) != 2:
            raise ValueError("visualize requires exactly 2 numeric features.")

        X = self.X.values
        y = self.y.values

        fig, ax = plt.subplots(figsize=figsize)
        for label in self.classes:
            mask = y == label
            ax.scatter(X[mask, 0], X[mask, 1], label=str(label), alpha=0.7, edgecolors='black')
        ax.set_xlabel(self.feature_names[0])
        ax.set_ylabel(self.feature_names[1])
        ax.set_title('KNN Classification (2D)')
        ax.legend()
        ax.grid(True, alpha=0.3)
        plt.tight_layout()
        plt.show()

    def summary(self, display='inverted'):
        self._check_fitted()
        y_pred = self.model.predict(self.X)
        report = self._compute_classification_metrics(self.y, y_pred)

        print("\n" + "=" * 70)
        print("KNN CLASSIFIER SUMMARY")
        print("=" * 70)
        print(f"Target: {self.target}")
        print(f"Classes: {self.classes}")
        print(f"Features: {len(self.feature_names)}")
        print(f"k: {self.k}")
        print(f"Weights: {self.weights}")
        print(f"Metric: {self.metric}")
        print(f"Training samples: {len(self.df)}")

        print("\nTRAINING FIT:")
        print("-" * 70)
        print(f"Training Accuracy: {report['accuracy']:.2%}")

        conf_obj = ConfusionMatrix.from_matrix(report['confusion_matrix'], labels=self.classes)
        conf_df = conf_obj.to_dataframe(display=display)
        print("\nTraining Confusion Matrix:")
        print(conf_df.to_string())
        print(f"\nKappa: {report['kappa']:.4f}")

        print("\n" + "=" * 70)


def fit_knn_classifier(df, formula, k=5, weights='uniform', metric='minkowski'):
    if formula is None:
        raise ValueError("Must provide 'formula' for model specification")
    return KNNClassifierModel(df, formula=formula, k=k, weights=weights, metric=metric)
